import json
from dataclasses import replace
from unittest.mock import MagicMock, patch

import pytest

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    DataCollectionGroupInfo,
    DataCollectionInfo,
    DataCollectionPositionInfo,
    Orientation,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)

from ......conftest import (
    EXPECTED_END_TIME,
    EXPECTED_START_TIME,
    TEST_BARCODE,
    TEST_SAMPLE_ID,
)
from ......expeye_helpers import (
    DC_COMMENT_RE,
    DC_RE,
    DCG_RE,
    DCGS_RE,
    DCS_RE,
    GRID_RE,
    POSITION_RE,
    TEST_DATA_COLLECTION_GROUP_ID,
    TEST_DATA_COLLECTION_IDS,
    TEST_GRID_INFO_IDS,
)

TEST_PROPOSAL_REF = "cm31105"
TEST_VISIT_NUMBER = 4


@pytest.fixture
def dummy_collection_group_info():
    return DataCollectionGroupInfo(
        visit_string="cm31105-4",
        experiment_type="Mesh3D",
        sample_id=364758,
    )


DC_INFO_FOR_BEGIN_XY = DataCollectionInfo(
    omega_start=0.0,
    data_collection_number=1,
    xtal_snapshot1="test_1_y",
    xtal_snapshot2="test_2_y",
    xtal_snapshot3="test_3_y",
    n_images=800,
    axis_range=0,
    axis_end=0.0,
    kappa_start=None,
    parent_id=None,
    visit_string="cm31105-4",
    ispyb_detector_id=78,
    axis_start=0.0,
    slitgap_vertical=0.1,
    slitgap_horizontal=0.1,
    beamsize_at_samplex=0.1,
    beamsize_at_sampley=0.1,
    transmission=100.0,
    comments="MX-Bluesky: Xray centring 1 -",
    detector_distance=100.0,
    exp_time=0.1,
    imgdir="/tmp/",
    file_template="file_name_0_master.h5",
    imgprefix="file_name",
    imgsuffix="h5",
    n_passes=1,
    overlap=0,
    start_image_number=1,
    wavelength=123.98419840550369,
    xbeam=150.0,
    ybeam=160.0,
    synchrotron_mode=None,
    undulator_gap1=1.0,
    start_time=EXPECTED_START_TIME,
)

DC_INFO_FOR_BEGIN_XZ = replace(
    DC_INFO_FOR_BEGIN_XY,
    xtal_snapshot1="test_1_z",
    xtal_snapshot2="test_2_z",
    xtal_snapshot3="test_3_z",
    omega_start=90.0,
    n_images=400,
    axis_end=90.0,
    axis_start=90.0,
    file_template="file_name_1_master.h5",
    comments="MX-Bluesky: Xray centring 2 -",
)

DC_INFO_FOR_UPDATE_XY = replace(
    DC_INFO_FOR_BEGIN_XY,
    parent_id=34,
    comments="Diffraction grid scan of 40 by 20 images in 100.0 um by 100.0 um steps. Top left (px): [50,100], bottom right (px): [3250,1700].",
    flux=10.0,
    synchrotron_mode="test",
)

DC_INFO_FOR_UPDATE_XZ = replace(
    DC_INFO_FOR_BEGIN_XZ,
    parent_id=34,
    comments="Diffraction grid scan of 40 by 10 images in 100.0 um by 200.0 um steps. Top left (px): [50,120], bottom right (px): [3250,1720].",
    flux=10.0,
    synchrotron_mode="test",
)

EXPECTED_BASE_UPSERT = {
    "detectorId": 78,
    "axisRange": 0,
    "slitGapVertical": 0.1,
    "slitGapHorizontal": 0.1,
    "beamSizeAtSampleX": 0.1,
    "beamSizeAtSampleY": 0.1,
    "transmission": 100.0,
    "dataCollectionNumber": 1,
    "detectorDistance": 100.0,
    "exposureTime": 0.1,
    "imageDirectory": "/tmp/",
    "imagePrefix": "file_name",
    "imageSuffix": "h5",
    "numberOfPasses": 1,
    "overlap": 0,
    "startImageNumber": 1,
    "wavelength": 123.98419840550369,
    "xBeam": 150.0,
    "yBeam": 160.0,
    "undulatorGap1": 1.0,
    "startTime": EXPECTED_START_TIME,
}

EXPECTED_BASE_XY_UPSERT = EXPECTED_BASE_UPSERT | {
    "xtalSnapshotFullPath1": "test_1_y",
    "xtalSnapshotFullPath2": "test_2_y",
    "xtalSnapshotFullPath3": "test_3_y",
    "omegaStart": 0,
    "axisStart": 0.0,
    "axisEnd": 0,
    "fileTemplate": "file_name_0_master.h5",
    "numberOfImages": 40 * 20,
}

EXPECTED_BASE_XZ_UPSERT = EXPECTED_BASE_UPSERT | {
    "xtalSnapshotFullPath1": "test_1_z",
    "xtalSnapshotFullPath2": "test_2_z",
    "xtalSnapshotFullPath3": "test_3_z",
    "omegaStart": 90.0,
    "axisStart": 90.0,
    "axisEnd": 90.0,
    "fileTemplate": "file_name_1_master.h5",
    "numberOfImages": 40 * 10,
}

EXPECTED_DC_XY_BEGIN_UPSERT = EXPECTED_BASE_XY_UPSERT | {
    "comments": "MX-Bluesky: Xray centring 1 -",
}

EXPECTED_DC_XZ_BEGIN_UPSERT = EXPECTED_BASE_XZ_UPSERT | {
    "comments": "MX-Bluesky: Xray centring 2 -",
}

EXPECTED_DC_XY_UPDATE_UPSERT = EXPECTED_BASE_XY_UPSERT | {
    "flux": 10.0,
    "synchrotronMode": "test",
}

EXPECTED_DC_XZ_UPDATE_UPSERT = EXPECTED_BASE_XZ_UPSERT | {
    "flux": 10,
    "synchrotronMode": "test",
}


@pytest.fixture
def scan_data_infos_for_begin():
    return [
        ScanDataInfo(
            data_collection_info=replace(DC_INFO_FOR_BEGIN_XY),
            data_collection_id=None,
            data_collection_position_info=None,
            data_collection_grid_info=None,
        ),
        ScanDataInfo(
            data_collection_info=replace(DC_INFO_FOR_BEGIN_XZ),
            data_collection_id=None,
            data_collection_position_info=None,
            data_collection_grid_info=None,
        ),
    ]


@pytest.fixture
def scan_data_infos_for_update():
    scan_xy_data_info_for_update = ScanDataInfo(
        data_collection_info=replace(DC_INFO_FOR_UPDATE_XY),
        data_collection_id=TEST_DATA_COLLECTION_IDS[0],
        data_collection_position_info=DataCollectionPositionInfo(
            pos_x=0, pos_y=0, pos_z=0
        ),
        data_collection_grid_info=DataCollectionGridInfo(
            dx_in_mm=0.1,
            dy_in_mm=0.1,
            steps_x=40,
            steps_y=20,
            microns_per_pixel_x=1.25,
            microns_per_pixel_y=1.25,
            snapshot_offset_x_pixel=50,
            snapshot_offset_y_pixel=100,
            orientation=Orientation.HORIZONTAL,
            snaked=True,
        ),
    )
    scan_xz_data_info_for_update = ScanDataInfo(
        data_collection_info=replace(DC_INFO_FOR_UPDATE_XZ),
        data_collection_id=TEST_DATA_COLLECTION_IDS[1],
        data_collection_position_info=DataCollectionPositionInfo(
            pos_x=0.0, pos_y=0.0, pos_z=0.0
        ),
        data_collection_grid_info=DataCollectionGridInfo(
            dx_in_mm=0.1,
            dy_in_mm=0.2,
            steps_x=40,
            steps_y=10,
            microns_per_pixel_x=1.25,
            microns_per_pixel_y=1.25,
            snapshot_offset_x_pixel=50,
            snapshot_offset_y_pixel=120,
            orientation=Orientation.HORIZONTAL,
            snaked=True,
        ),
    )
    return [scan_xy_data_info_for_update, scan_xz_data_info_for_update]


def test_ispyb_deposition_comment_for_3d_correct(
    mock_ispyb_conn: MagicMock,
    dummy_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    mock_ispyb_conn = mock_ispyb_conn

    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)

    first_create_dc, second_create_dc = mock_ispyb_conn.dc_calls_for(DCS_RE)
    assert first_create_dc.body["comments"] == ("MX-Bluesky: Xray centring 1 -")
    assert second_create_dc.body["comments"] == ("MX-Bluesky: Xray centring 2 -")
    append_comments_requests = mock_ispyb_conn.dc_calls_for(DC_COMMENT_RE)
    assert append_comments_requests[0].dcid == TEST_DATA_COLLECTION_IDS[0]
    assert append_comments_requests[1].dcid == TEST_DATA_COLLECTION_IDS[1]
    assert (
        append_comments_requests[0].body["comments"]
        == " Diffraction grid scan of 40 by 20 images "
        "in 100.0 um by 100.0 um steps. Top left (px): [50,100], bottom right (px): [3250,1700]."
    )
    assert (
        append_comments_requests[1].body["comments"]
        == " Diffraction grid scan of 40 by 10 images "
        "in 100.0 um by 200.0 um steps. Top left (px): [50,120], bottom right (px): [3250,1720]."
    )


def test_store_3d_grid_scan(
    mock_ispyb_conn,
    dummy_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    assert ispyb_ids == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0], TEST_DATA_COLLECTION_IDS[1]),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )

    assert dummy_ispyb.update_deposition(
        ispyb_ids, scan_data_infos_for_update
    ) == IspybIds(
        data_collection_ids=TEST_DATA_COLLECTION_IDS,
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        grid_ids=TEST_GRID_INFO_IDS,
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_begin_deposition(
    mock_ispyb_conn,
    dummy_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
):
    assert dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0], TEST_DATA_COLLECTION_IDS[1]),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )

    create_dcg_request = mock_ispyb_conn.calls_for(DCGS_RE)[0].request
    assert mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 2) == TEST_PROPOSAL_REF
    assert (
        int(mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 3)) == TEST_VISIT_NUMBER
    )
    dcg_payload = json.loads(create_dcg_request.body)
    assert dcg_payload["experimentType"] == "Mesh3D"
    assert dcg_payload["sampleId"] == TEST_SAMPLE_ID

    create_dc_requests = mock_ispyb_conn.calls_for(DCS_RE)
    request1 = create_dc_requests[0].request
    assert (
        int(mock_ispyb_conn.match(request1, DCS_RE, 2)) == TEST_DATA_COLLECTION_GROUP_ID
    )
    assert json.loads(request1.body) == EXPECTED_DC_XY_BEGIN_UPSERT

    request2 = create_dc_requests[1].request
    assert (
        int(mock_ispyb_conn.match(request2, DCS_RE, 2)) == TEST_DATA_COLLECTION_GROUP_ID
    )
    assert json.loads(request2.body) == EXPECTED_DC_XZ_BEGIN_UPSERT

    assert len(mock_ispyb_conn.calls_for(POSITION_RE)) == 0
    assert len(mock_ispyb_conn.calls_for(GRID_RE)) == 0


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_update_deposition(
    mock_ispyb_conn,
    dummy_ispyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    assert len(mock_ispyb_conn.calls_for(DCGS_RE)) == 1
    assert len(mock_ispyb_conn.calls_for(DCS_RE)) == 2

    dummy_collection_group_info.sample_barcode = TEST_BARCODE

    actual_rows = dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)

    assert actual_rows == IspybIds(
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        data_collection_ids=TEST_DATA_COLLECTION_IDS,
        grid_ids=TEST_GRID_INFO_IDS,
    )

    assert len(mock_ispyb_conn.calls_for(DCGS_RE)) == 1

    update_xy_req, update_xz_req = mock_ispyb_conn.dc_calls_for(DC_RE)

    assert update_xy_req.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_xy_req.body == EXPECTED_DC_XY_UPDATE_UPSERT
    assert update_xz_req.dcid == TEST_DATA_COLLECTION_IDS[1]
    assert update_xz_req.body == EXPECTED_DC_XZ_UPDATE_UPSERT

    dcid_to_comment_req = {
        rq.dcid: rq for rq in mock_ispyb_conn.dc_calls_for(DC_COMMENT_RE)
    }
    assert dcid_to_comment_req[TEST_DATA_COLLECTION_IDS[0]].body == {
        "comments": " Diffraction grid scan of 40 by 20 "
        "images in 100.0 um by 100.0 um steps. Top left (px): [50,100], "
        "bottom right (px): [3250,1700].",
    }

    update_dc_xy_pos_req, update_dc_xz_pos_req = mock_ispyb_conn.dc_calls_for(
        POSITION_RE
    )
    assert update_dc_xy_pos_req.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_xz_pos_req.body == {
        "posX": 0,
        "posY": 0,
        "posZ": 0,
    }

    update_grid_xy_req, update_grid_xz_req = mock_ispyb_conn.dc_calls_for(GRID_RE)
    assert update_grid_xy_req.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_grid_xy_req.body == {
        "dx": 0.1,
        "dy": 0.1,
        "stepsX": 40,
        "stepsY": 20,
        "micronsPerPixelX": 1.25,
        "micronsPerPixelY": 1.25,
        "snapshotOffsetXPixel": 50,
        "snapshotOffsetYPixel": 100,
        "orientation": "horizontal",
        "snaked": True,
    }

    assert dcid_to_comment_req[TEST_DATA_COLLECTION_IDS[1]].body == {
        "comments": " Diffraction grid scan of 40 by 10 "
        "images in 100.0 um by 200.0 um steps. Top left (px): [50,120], "
        "bottom right (px): [3250,1720].",
    }

    assert update_dc_xz_pos_req.dcid == TEST_DATA_COLLECTION_IDS[1]
    assert update_dc_xz_pos_req.body == {
        "posX": 0,
        "posY": 0,
        "posZ": 0,
    }

    assert update_grid_xz_req.dcid == TEST_DATA_COLLECTION_IDS[1]
    assert update_grid_xz_req.body == {
        "dx": 0.1,
        "dy": 0.2,
        "stepsX": 40,
        "stepsY": 10,
        "micronsPerPixelX": 1.25,
        "micronsPerPixelY": 1.25,
        "snapshotOffsetXPixel": 50,
        "snapshotOffsetYPixel": 120,
        "orientation": "horizontal",
        "snaked": True,
    }


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
@patch(
    "mx_bluesky.common.external_interaction.ispyb.ispyb_store.get_current_time_string",
)
def test_end_deposition_happy_path(
    get_current_time,
    mock_ispyb_conn,
    dummy_ispyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    assert len(mock_ispyb_conn.calls_for(DCGS_RE)) == 1
    ispyb_ids = dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)
    assert len(mock_ispyb_conn.calls_for(DCGS_RE)) == 1
    assert len(mock_ispyb_conn.calls_for(DCS_RE)) == 2
    assert len(mock_ispyb_conn.calls_for(DC_RE)) == 2
    assert len(mock_ispyb_conn.calls_for(GRID_RE)) == 2

    get_current_time.return_value = EXPECTED_END_TIME
    dummy_ispyb.end_deposition(ispyb_ids, "success", "Test succeeded")
    dcids_to_append_comment_reqs = {
        rq.dcid: rq for rq in mock_ispyb_conn.dc_calls_for(DC_COMMENT_RE)
    }
    assert dcids_to_append_comment_reqs[TEST_DATA_COLLECTION_IDS[0]].body == {
        "comments": " DataCollection Successful reason: Test succeeded"
    }

    update_dc_requests = list(mock_ispyb_conn.dc_calls_for(DC_RE)[2:])
    assert update_dc_requests[0].dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_requests[0].body == {
        "endTime": EXPECTED_END_TIME,
        "runStatus": "DataCollection Successful",
    }
    assert dcids_to_append_comment_reqs[TEST_DATA_COLLECTION_IDS[1]].body == {
        "comments": " DataCollection Successful reason: Test succeeded",
    }
    assert update_dc_requests[1].dcid == TEST_DATA_COLLECTION_IDS[1]
    assert update_dc_requests[1].body == {
        "endTime": EXPECTED_END_TIME,
        "runStatus": "DataCollection Successful",
    }


def test_param_keys(
    mock_ispyb_conn,
    dummy_ispyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    assert dummy_ispyb.update_deposition(
        ispyb_ids, scan_data_infos_for_update
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0], TEST_DATA_COLLECTION_IDS[1]),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        grid_ids=(TEST_GRID_INFO_IDS[0], TEST_GRID_INFO_IDS[1]),
    )


def test_given_sampleid_of_none_when_grid_scan_stored_then_sample_id_not_set(
    mock_ispyb_conn,
    dummy_ispyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    dummy_collection_group_info.sample_id = None
    for dc_info in [
        scan_info.data_collection_info
        for scan_info in scan_data_infos_for_begin + scan_data_infos_for_update
    ]:
        dc_info.sample_id = None

    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)

    dcg_requests = [c.request for c in mock_ispyb_conn.calls_for(DCGS_RE)] + [
        c.request for c in mock_ispyb_conn.calls_for(DCG_RE)
    ]
    for req in dcg_requests:
        assert "sampleId" not in json.loads(req.body)


def test_given_real_sampleid_when_grid_scan_stored_then_sample_id_set(
    mock_ispyb_conn,
    dummy_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    expected_sample_id = 364758

    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)

    dcg_requests = [c.request for c in mock_ispyb_conn.calls_for(DCGS_RE)] + [
        c.request for c in mock_ispyb_conn.calls_for(DCG_RE)
    ]
    assert len(dcg_requests) > 0
    for req in dcg_requests:
        assert json.loads(req.body)["sampleId"] == expected_sample_id


def test_fail_result_run_results_in_bad_run_status(
    mock_ispyb_conn: MagicMock,
    dummy_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    mock_ispyb_conn = mock_ispyb_conn

    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    ispyb_ids = dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)
    dummy_ispyb.end_deposition(ispyb_ids, "fail", "test specifies failure")

    update_dc_requests = list(mock_ispyb_conn.dc_calls_for(DC_RE))
    expected_ids_and_runstatuses = [
        (TEST_DATA_COLLECTION_IDS[0], None),
        (TEST_DATA_COLLECTION_IDS[1], None),
        (TEST_DATA_COLLECTION_IDS[0], "DataCollection Unsuccessful"),
        (TEST_DATA_COLLECTION_IDS[1], "DataCollection Unsuccessful"),
    ]
    for req, expected in zip(
        update_dc_requests, expected_ids_and_runstatuses, strict=True
    ):
        assert req.dcid == expected[0]
        assert req.body.get("runStatus") == expected[1]


def test_fail_result_long_comment_still_updates_run_status(
    mock_ispyb_conn: MagicMock,
    dummy_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    mock_ispyb_conn = mock_ispyb_conn

    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    ispyb_ids = dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)
    dummy_ispyb.end_deposition(ispyb_ids, "fail", "this comment is too long")

    update_dc_requests = list(mock_ispyb_conn.dc_calls_for(DC_RE))
    expected_ids_and_runstatuses = [
        (TEST_DATA_COLLECTION_IDS[0], None),
        (TEST_DATA_COLLECTION_IDS[1], None),
        (TEST_DATA_COLLECTION_IDS[0], "DataCollection Unsuccessful"),
        (TEST_DATA_COLLECTION_IDS[1], "DataCollection Unsuccessful"),
    ]
    for req, expected in zip(
        update_dc_requests, expected_ids_and_runstatuses, strict=True
    ):
        assert req.dcid == expected[0]
        assert req.body.get("runStatus") == expected[1]


def test_no_exception_during_run_results_in_good_run_status(
    mock_ispyb_conn: MagicMock,
    dummy_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    ispyb_ids = dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)
    dummy_ispyb.end_deposition(ispyb_ids, "success", "")

    update_dc_requests = [c.request for c in mock_ispyb_conn.calls_for(DC_RE)]
    expected_ids_and_runstatuses = [
        (TEST_DATA_COLLECTION_IDS[0], None),
        (TEST_DATA_COLLECTION_IDS[1], None),
        (TEST_DATA_COLLECTION_IDS[0], "DataCollection Successful"),
        (TEST_DATA_COLLECTION_IDS[1], "DataCollection Successful"),
    ]
    for req, expected in zip(
        update_dc_requests, expected_ids_and_runstatuses, strict=True
    ):
        payload = json.loads(req.body)
        assert int(mock_ispyb_conn.match(req, DC_RE, 2)) == expected[0]
        assert payload.get("runStatus") == expected[1]


def test_update_data_collection_no_comment(
    mock_ispyb_conn: MagicMock,
    dummy_ispyb: StoreInIspyb,
    dummy_collection_group_info: DataCollectionGroupInfo,
    scan_data_infos_for_begin: list[ScanDataInfo],
    scan_data_infos_for_update: list[ScanDataInfo],
):
    for scan_data_info in scan_data_infos_for_update:
        scan_data_info.data_collection_info.comments = None

    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)

    assert len(mock_ispyb_conn.calls_for(DC_COMMENT_RE)) == 0



from __future__ import annotations

import json

from dodal.devices.oav.oav_parameters import OAVParameters

from mx_bluesky.common.device_setup_plans.manipulate_sample import move_phi_chi_omega
from mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan import (
    detect_grid_and_do_gridscan,
)
from mx_bluesky.common.experiment_plans.inner_plans.xrc_results_utils import (
    fetch_xrc_results_from_zocalo,
    zocalo_stage_decorator,
)
from mx_bluesky.common.experiment_plans.oav_snapshot_plan import (
    setup_beamline_for_oav,
)
from mx_bluesky.common.experiment_plans.pin_tip_centring_plan import (
    PinTipCentringComposite,
    pin_tip_centre_plan,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    ispyb_activation_wrapper,
)
from mx_bluesky.common.parameters.constants import OavConstants, PlanNameConstants
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.common.preprocessors.preprocessors import (
    pause_xbpm_feedback_during_collection_at_desired_transmission_decorator,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.experiment_plans.hyperion_flyscan_xray_centre_plan import (
    construct_hyperion_specific_features,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionGridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    GridScanWithEdgeDetect,
    HyperionSpecifiedThreeDGridScan,
    PinTipCentreThenXrayCentre,
)


def create_parameters_for_grid_detection(
    pin_centre_parameters: PinTipCentreThenXrayCentre,
) -> GridScanWithEdgeDetect:
    params_json = json.loads(pin_centre_parameters.model_dump_json())
    del params_json["tip_offset_um"]
    grid_detect_and_gridscan = GridScanWithEdgeDetect(**params_json)
    LOGGER.info(
        f"Parameters for grid detect and gridscan: {grid_detect_and_gridscan.model_dump_json(indent=2)}"
    )
    return grid_detect_and_gridscan


def pin_centre_then_gridscan_plan(
    composite: HyperionGridDetectThenXRayCentreComposite,
    parameters: PinTipCentreThenXrayCentre,
    oav_config_file: str = OavConstants.OAV_CONFIG_JSON,
):
    """Plan that performs a pin tip centre followed by a gridscan to determine the centres of interest"""

    pin_tip_centring_composite = PinTipCentringComposite(
        oav=composite.oav,
        gonio=composite.gonio,
        pin_tip_detection=composite.pin_tip_detection,
    )

    @zocalo_stage_decorator(composite.zocalo)
    def _pin_centre_then_gridscan_and_xrc():
        yield from setup_beamline_for_oav(
            composite.gonio, composite.backlight, composite.aperture_scatterguard
        )

        yield from move_phi_chi_omega(
            composite.gonio,
            parameters.phi_start_deg,
            parameters.chi_start_deg,
            group=CONST.WAIT.READY_FOR_OAV,
        )

        yield from pin_tip_centre_plan(
            pin_tip_centring_composite,
            parameters.tip_offset_um,
            oav_config_file,
        )

        grid_detect_params = create_parameters_for_grid_detection(parameters)
        oav_params = OAVParameters("xrayCentring", oav_config_file)

        @pause_xbpm_feedback_during_collection_at_desired_transmission_decorator(
            composite,
            parameters.transmission_frac,
            PlanNameConstants.GRIDSCAN_OUTER,
        )
        def _grid_detect_and_gridscan_plan():
            yield from detect_grid_and_do_gridscan(
                composite,
                grid_detect_params,
                oav_params,
                HyperionSpecifiedThreeDGridScan,
                construct_hyperion_specific_features,
            )

        yield from _grid_detect_and_gridscan_plan()
        assert isinstance(
            grid_detect_params.specified_grid_params, SpecifiedThreeDGridScan
        ), "Specified grid params couldn't be found after grid detection"
        yield from fetch_xrc_results_from_zocalo(
            composite.zocalo, grid_detect_params.specified_grid_params
        )

    yield from ispyb_activation_wrapper(_pin_centre_then_gridscan_and_xrc(), parameters)



import asyncio
from collections.abc import Generator
from concurrent.futures import Executor, Future, ThreadPoolExecutor
from importlib import resources
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest
from dodal.beamlines import i03

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGroupInfo,
)
from mx_bluesky.common.parameters.rotation import (
    RotationScan,
)
from mx_bluesky.hyperion._plan_runner_params import Wait
from mx_bluesky.hyperion.blueapi.parameters import LoadCentreCollectParams
from mx_bluesky.hyperion.parameters.gridscan import (
    GridScanWithEdgeDetect,
    HyperionSpecifiedThreeDGridScan,
)
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.plan_runner import PlanRunner
from tests.conftest import (
    raw_params_from_file,
)

i03.DAQ_CONFIGURATION_PATH = "tests/test_data/test_daq_configuration"
BANNED_PATHS = [Path("/dls"), Path("/dls_sw")]

# Time to wait for the whole test script thread to complete
TEST_SCRIPT_TIMEOUT_S = 2

# Time to wait for the test to progress to the next step
AGAMEMNON_WAIT_FOR_TEST_STEP_S = 0.2

AGAMEMNON_WAIT_INSTRUCTION = Wait.model_validate(
    {
        "duration_s": AGAMEMNON_WAIT_FOR_TEST_STEP_S,
    }
)


@pytest.fixture(autouse=True)
def override_hyperion_blueapi_logging_setup(request):
    log_path = Path("/tmp/logs/bluesky")
    with patch(
        "mx_bluesky.common.utils.log._get_logging_dirs",
        return_value=(log_path, log_path),
    ):
        yield


@pytest.fixture(scope="session")
def executor() -> Generator[Executor, Any, Any]:
    ex = ThreadPoolExecutor(max_workers=1, thread_name_prefix="test thread")
    yield ex
    ex.shutdown(wait=True)


@pytest.fixture
def load_centre_collect_params(tmp_path) -> LoadCentreCollect:
    json_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_load_centre_collect_params.json",
        tmp_path,
    )
    return LoadCentreCollect(**json_dict)


@pytest.fixture
def external_load_centre_collect_params(tmp_path) -> LoadCentreCollectParams:
    json_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/external_load_centre_collect_params.json",
        tmp_path,
    )
    return LoadCentreCollectParams(**json_dict)


@pytest.fixture(autouse=True)
def patch_open_to_prevent_dls_reads_in_tests():
    unpatched_open = open
    assert __package__
    project_folder = resources.files(__package__)
    assert isinstance(project_folder, Path)
    project_folder = project_folder.parent.parent.parent

    def patched_open(*args, **kwargs):
        requested_path = Path(args[0])
        if requested_path.is_absolute():
            for p in BANNED_PATHS:
                assert not requested_path.is_relative_to(p), (
                    f"Attempt to open {requested_path} from inside a unit test"
                )
        return unpatched_open(*args, **kwargs)

    with patch("builtins.open", side_effect=patched_open):
        yield []


@pytest.fixture
def test_rotation_params_nomove(tmp_path):
    return RotationScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_one_multi_rotation_scan_parameters_nomove.json",
            tmp_path,
        )
    )


@pytest.fixture
def test_multi_rotation_params(tmp_path):
    return RotationScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_multi_rotation_scan_parameters.json",
            tmp_path,
        )
    )


@pytest.fixture
def test_fgs_params(tmp_path):
    return HyperionSpecifiedThreeDGridScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_parameters.json", tmp_path
        )
    )


@pytest.fixture(params=[False, True])
def test_omega_flip(request):
    with patch(
        "mx_bluesky.common.parameters.constants.RotationParamConstants.OMEGA_FLIP",
        new=request.param,
    ):
        yield request.param


@pytest.fixture
def fgs_params_use_panda(tmp_path):
    with patch(
        "mx_bluesky.hyperion.external_interaction.config_server.GDA_DOMAIN_PROPERTIES_PATH",
        new="tests/test_data/test_domain_properties_with_panda",
    ):
        params = raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_parameters.json",
            tmp_path,
        )
        yield HyperionSpecifiedThreeDGridScan(**params)


@pytest.fixture
def test_full_grid_scan_params(tmp_path):
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_grid_with_edge_detect_parameters.json",
        tmp_path,
    )
    return GridScanWithEdgeDetect(**params)


def dummy_rotation_data_collection_group_info():
    return DataCollectionGroupInfo(
        visit_string="cm31105-4",
        experiment_type="SAD",
        sample_id=364758,
    )


def launch_test_in_runner_event_loop(
    async_func, udc_runner: PlanRunner, executor
) -> Future:
    """Launch the async func in a separate thread because the RunEngine under
    test must run in the main thread and block our test code, and return
    result and any exception to the caller."""

    def _launch_in_new_thread():
        future = asyncio.run_coroutine_threadsafe(
            async_func(), udc_runner.run_engine.loop
        )
        return future.result(TEST_SCRIPT_TIMEOUT_S)

    return executor.submit(_launch_in_new_thread)


@pytest.fixture()
def use_beamline_i03(monkeypatch):
    monkeypatch.setenv("BEAMLINE", "i03")



"""
This module is the bluesky plan module for use with hyperion-blueapi.
Importing this module will configure debug and info logging as a side-effect - so this module should not be
imported directly by other components as it is intended only as the entry-point for BlueAPI.
"""

from mx_bluesky.common.utils.log import setup_hyperion_blueapi_logging
from mx_bluesky.hyperion.blueapi.in_process import (
    clean_up_udc,
    load_centre_collect,
    move_to_udc_default_state,
    robot_unload,
)
from mx_bluesky.hyperion.blueapi.parameters import (
    LoadCentreCollectParams,
)
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    LoadCentreCollectComposite,
)
from mx_bluesky.hyperion.experiment_plans.udc_default_state import (
    UDCDefaultDevices,
)
from mx_bluesky.hyperion.parameters.constants import CONST

__all__ = [
    "LoadCentreCollectComposite",
    "LoadCentreCollectParams",
    "UDCDefaultDevices",
    "clean_up_udc",
    "load_centre_collect",
    "move_to_udc_default_state",
    "robot_unload",
]


def _init_plan_module():
    """Initialisation hooks for hyperion-blueapi"""
    setup_hyperion_blueapi_logging(CONST.LOG_FILE_NAME)


_init_plan_module()



from functools import partial
from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest
from bluesky import Msg
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import MsgGenerator
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.beamlines.i04.beamsize import Beamsize
from dodal.devices.beamlines.i04.transfocator import Transfocator
from dodal.devices.common_dcm import DoubleCrystalMonochromator
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import (
    ZebraFastGridScanThreeD,
)
from dodal.devices.flux import Flux
from dodal.devices.mx_phase1.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.undulator import UndulatorInKeV
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from dodal.devices.zocalo.zocalo_results import (
    ZOCALO_STAGE_GROUP,
)
from ophyd_async.core import get_mock_put, set_mock_value

from mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan import (
    DEFAULT_XRC_BEAMSIZE_MICRONS,
    I04AutoXrcParams,
    _get_grid_common_params,
    get_ready_for_oav_and_close_shutter,
    i04_default_grid_detect_and_xray_centre,
)
from mx_bluesky.common.parameters.constants import PlanNameConstants
from mx_bluesky.common.parameters.gridscan import (
    GridCommon,
    SpecifiedThreeDGridScan,
)
from mx_bluesky.common.utils.exceptions import CrystalNotFoundError
from mx_bluesky.common.utils.xrc_result import XRayCentreResult
from tests.conftest import TEST_RESULT_LARGE, fake_generator, simulate_xrc_result
from tests.unit_tests.common.experiment_plans.test_common_flyscan_xray_centre_plan import (
    CompleteError,
)


class CustomError(Exception): ...


EXPECTED_WAVELENGTH = 0.95373


@pytest.fixture
def i04_grid_detect_then_xrc_default_params(
    aperture_scatterguard: ApertureScatterguard,
    attenuator: BinaryFilterAttenuator,
    backlight: Backlight,
    beamsize: Beamsize,
    beamstop_phase1: Beamstop,
    dcm: DoubleCrystalMonochromator,
    zebra_fast_grid_scan: ZebraFastGridScanThreeD,
    flux: Flux,
    oav: OAV,
    pin_tip_detection_with_found_pin: PinTipDetection,
    s4_slit_gaps: S4SlitGaps,
    undulator: UndulatorInKeV,
    xbpm_feedback: XBPMFeedback,
    zebra: Zebra,
    robot: BartRobot,
    sample_shutter: ZebraShutter,
    eiger: EigerDetector,
    synchrotron: Synchrotron,
    zocalo: ZocaloResults,
    smargon: Smargon,
    detector_motion: DetectorMotion,
    transfocator: Transfocator,
    tmp_path,
):
    entry_params = I04AutoXrcParams(
        sample_id=1,
        file_name="filename",
        visit="cm40607-5",
        detector_distance_mm=264.5,
        storage_directory=str(tmp_path),
    )

    set_mock_value(dcm.wavelength_in_a.user_readback, EXPECTED_WAVELENGTH)
    return partial(
        i04_default_grid_detect_and_xray_centre,
        parameters=entry_params,
        aperture_scatterguard=aperture_scatterguard,
        attenuator=attenuator,
        backlight=backlight,
        beamstop=beamstop_phase1,
        beamsize=beamsize,
        dcm=dcm,
        zebra_fast_grid_scan=zebra_fast_grid_scan,
        flux=flux,
        oav=oav,
        pin_tip_detection=pin_tip_detection_with_found_pin,
        s4_slit_gaps=s4_slit_gaps,
        undulator=undulator,
        xbpm_feedback=xbpm_feedback,
        zebra=zebra,
        robot=robot,
        sample_shutter=sample_shutter,
        eiger=eiger,
        synchrotron=synchrotron,
        zocalo=zocalo,
        smargon=smargon,
        detector_motion=detector_motion,
        transfocator=transfocator,
    )


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.setup_beamline_for_oav",
    autospec=True,
)
def test_get_ready_for_oav_and_close_shutter_closes_shutter_and_calls_setup_for_oav_plan(
    mock_setup_beamline_for_oav: MagicMock,
    sim_run_engine: RunEngineSimulator,
    grid_detect_xrc_devices,
):
    mock_setup_beamline_for_oav.return_value = iter([Msg("setup_beamline_for_oav")])

    msgs = sim_run_engine.simulate_plan(
        get_ready_for_oav_and_close_shutter(
            grid_detect_xrc_devices.gonio,
            grid_detect_xrc_devices.backlight,
            grid_detect_xrc_devices.aperture_scatterguard,
            grid_detect_xrc_devices.detector_motion,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs, predicate=lambda msg: msg.command == "wait"
    )

    msgs = assert_message_and_return_remaining(
        msgs, predicate=lambda msg: msg.command == "setup_beamline_for_oav"
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "set"
        and msg.obj.name == "detector_motion-shutter"
        and msg.args[0] == 0,
    )
    msgs = assert_message_and_return_remaining(
        msgs, predicate=lambda msg: msg.command == "wait"
    )


@pytest.mark.parametrize(
    "udc",
    [(True), (False)],
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_ready_for_oav_and_close_shutter",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.grid_detect_then_xray_centre",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.setup_beamline_for_oav",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.create_gridscan_callbacks",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_results_and_move_to_xtal",
    new=MagicMock(),
)
def test_i04_grid_detect_then_xrc_closes_shutter_and_tidies_if_not_udc(
    mock_create_gridscan_callbacks: MagicMock,
    mock_setup_beamline_for_oav: MagicMock,
    mock_grid_detect_then_xray_centre: MagicMock,
    mock_get_ready_for_oav_and_close_shutter: MagicMock,
    run_engine: RunEngine,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
    udc: bool,
    give_grid_common_specific_grid,
):
    run_engine(
        i04_grid_detect_then_xrc_default_params(
            udc=udc,
        )
    )

    call_count = 0 if udc else 1

    assert mock_get_ready_for_oav_and_close_shutter.call_count == call_count


@patch("bluesky.plan_stubs.sleep", autospec=True)
@patch(
    "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.fix_transmission_and_exposure_time_for_current_wavelength"
)
@patch(
    "mx_bluesky.common.preprocessors.preprocessors.check_and_pause_feedback",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_results_and_move_to_xtal",
    new=MagicMock(),
)
def test_i04_default_grid_detect_and_xray_centre_sets_transmission_and_triggers_xbpm_feedback_before_run(
    mock_pause_feedback: MagicMock,
    mock_fix_transmission_and_exp_time: MagicMock,
    mock_create_parameters: MagicMock,
    mock_grid_detection_callback: MagicMock,
    mock_grid_detection_plan: MagicMock,
    mock_check_topup: MagicMock,
    mock_wait: MagicMock,
    sim_run_engine: RunEngineSimulator,
    zocalo: ZocaloResults,
    hyperion_fgs_params,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
):
    desired_transmission = 0.4
    mock_fix_transmission_and_exp_time.return_value = (desired_transmission, 1)
    mock_create_parameters.return_value = hyperion_fgs_params
    simulate_xrc_result(
        sim_run_engine,
        zocalo,
        TEST_RESULT_LARGE,
    )

    msgs = sim_run_engine.simulate_plan(
        i04_grid_detect_then_xrc_default_params(),
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args == (desired_transmission,),
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "trigger" and msg.obj.name == "xbpm_feedback",
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run"
        and msg.run == PlanNameConstants.GRIDSCAN_OUTER,
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "close_run"
        and msg.run == PlanNameConstants.GRIDSCAN_OUTER,
    )

    mock_pause_feedback.assert_not_called()


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.create_gridscan_callbacks",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.move_aperture_if_required",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
)
@patch(
    "dodal.plans.preprocessors.verify_undulator_gap.verify_undulator_gap",
)
def test_i04_default_grid_detect_and_xray_centre_does_undulator_check_before_collection(
    mock_verify_gap: MagicMock,
    mock_run_gridscan: MagicMock,
    mock_create_parameters: MagicMock,
    mock_grid_params_callback: MagicMock,
    mock_move_aperture_if_required: MagicMock,
    mock_grid_detection_plan: MagicMock,
    mock_create_gridscan_callbacks: MagicMock,
    run_engine: RunEngine,
    hyperion_fgs_params,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
):
    mock_create_parameters.return_value = hyperion_fgs_params
    mock_run_gridscan.side_effect = CompleteError
    with pytest.raises(CompleteError):
        run_engine(i04_grid_detect_then_xrc_default_params())

    mock_verify_gap.assert_called_once()


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_ready_for_oav_and_close_shutter",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.create_gridscan_callbacks",
    autospec=True,
)
def test_i04_grid_detect_then_xrc_tidies_up_on_exception(
    mock_create_gridscan_callbacks: MagicMock,
    mock_get_ready_for_oav_and_close_shutter: MagicMock,
    run_engine: RunEngine,
    i04_grid_detect_then_xrc_default_params,
):
    mock_create_gridscan_callbacks.side_effect = CustomError
    with pytest.raises(CustomError):
        run_engine(
            i04_grid_detect_then_xrc_default_params(
                udc=False,
            )
        )

    assert mock_get_ready_for_oav_and_close_shutter.call_count == 1


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_ready_for_oav_and_close_shutter",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.grid_detect_then_xray_centre",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.create_gridscan_callbacks",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_results_and_move_to_xtal",
    new=MagicMock(),
)
async def test_i04_grid_detect_then_xrc_sets_beamsize_before_grid_detect_then_reverts(
    mock_create_gridscan_callbacks: MagicMock,
    mock_grid_detect_then_xray_centre: MagicMock,
    mock_get_ready_for_oav_and_close_shutter: MagicMock,
    run_engine: RunEngine,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
    transfocator: Transfocator,
    give_grid_common_specific_grid,
):
    initial_beamsize = 5.6
    set_mock_value(transfocator.current_vertical_size_rbv, initial_beamsize)
    parent_mock = MagicMock()
    assert isinstance(transfocator.set, MagicMock)
    parent_mock.attach_mock(transfocator.set, "transfocator_set")
    parent_mock.attach_mock(
        mock_create_gridscan_callbacks, "mock_create_gridscan_callbacks"
    )
    run_engine(i04_grid_detect_then_xrc_default_params())

    assert (
        mock_grid_detect_then_xray_centre.call_args.kwargs[
            "parameters"
        ].selected_aperture
        == ApertureValue.LARGE
    )
    assert parent_mock.method_calls == [
        call.transfocator_set(DEFAULT_XRC_BEAMSIZE_MICRONS),
        call.mock_create_gridscan_callbacks(),
        call.transfocator_set(initial_beamsize),
    ]


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_ready_for_oav_and_close_shutter",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.grid_detect_then_xray_centre",
    autospec=True,
)
async def test_given_no_diffraction_found_i04_grid_detect_then_xrc_returns_sample_to_initial_position(
    mock_grid_detect_then_xray_centre: MagicMock,
    mock_get_ready_for_oav_and_close_shutter: MagicMock,
    run_engine: RunEngine,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
    smargon: Smargon,
):
    initial_x, initial_y, initial_z = 1, 2, 3
    set_mock_value(smargon.x.user_readback, initial_x)
    set_mock_value(smargon.y.user_readback, initial_y)
    set_mock_value(smargon.z.user_readback, initial_z)

    mock_grid_detect_then_xray_centre.side_effect = CrystalNotFoundError

    with pytest.raises(CrystalNotFoundError):
        run_engine(i04_grid_detect_then_xrc_default_params())

    get_mock_put(smargon.x.user_setpoint).assert_has_calls([call(initial_x)])
    get_mock_put(smargon.y.user_setpoint).assert_has_calls([call(initial_y)])
    get_mock_put(smargon.z.user_setpoint).assert_has_calls([call(initial_z)])


@pytest.fixture()
def give_grid_common_specific_grid(test_fgs_params: SpecifiedThreeDGridScan):
    # Using this fixture means grid common always comes with its specific grid params
    class GridCommonWithSpecificGrid(GridCommon):
        @property
        def specified_grid_params(self):
            return test_fgs_params

    with patch(
        "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.GridCommon",
        GridCommonWithSpecificGrid,
    ):
        yield


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_ready_for_oav_and_close_shutter",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.grid_detect_then_xray_centre",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.setup_beamline_for_oav",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.create_gridscan_callbacks",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.fix_transmission_and_exposure_time_for_current_wavelength",
    return_value=(1, 0.004),
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_results_and_move_to_xtal",
    new=MagicMock(),
)
def test_i04_grid_detect_then_xrc_calculates_exposure_and_transmission_then_uses_grid_common(
    mock_fix_transmission: MagicMock,
    mock_create_gridscan_callbacks: MagicMock,
    mock_setup_beamline_for_oav: MagicMock,
    mock_grid_detect_then_xray_centre: MagicMock,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
    run_engine: RunEngine,
    test_full_grid_scan_params: GridCommon,
    test_fgs_params: SpecifiedThreeDGridScan,
    give_grid_common_specific_grid,
):
    expected_trans_frac = 1
    expected_exposure_time = 0.004
    mock_fix_transmission.return_value = (expected_trans_frac, expected_exposure_time)

    run_engine(i04_grid_detect_then_xrc_default_params())
    mock_fix_transmission.assert_called_once()

    grid_common_params = mock_grid_detect_then_xray_centre.call_args.kwargs[
        "parameters"
    ]
    assert isinstance(grid_common_params, GridCommon)
    assert grid_common_params.exposure_time_s == expected_exposure_time
    assert grid_common_params.transmission_frac == expected_trans_frac


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.fix_transmission_and_exposure_time_for_current_wavelength",
)
def test_get_grid_common_params(
    mock_fix_trans_and_exposure: MagicMock,
    tmp_path,
):
    expected_trans_frac = 0.2
    expected_exposure_time = 0.007
    mock_fix_trans_and_exposure.return_value = (
        expected_trans_frac,
        expected_exposure_time,
    )
    entry_params = I04AutoXrcParams(
        sample_id=1,
        file_name="filename",
        visit="cm40607-5",
        detector_distance_mm=264.5,
        storage_directory=str(tmp_path),
    )
    grid_common_params = _get_grid_common_params(1, entry_params)
    assert grid_common_params.exposure_time_s == expected_exposure_time
    assert grid_common_params.transmission_frac == expected_trans_frac


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.grid_detect_then_xray_centre",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.fix_transmission_and_exposure_time_for_current_wavelength",
    new=MagicMock(return_value=(1, 0.004)),
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_results_and_move_to_xtal",
    new=MagicMock(),
)
def test_grid_detect_then_xrc_stages_and_unstages_zocalo_and_gets_results(
    sim_run_engine: RunEngineSimulator,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
    give_grid_common_specific_grid,
):
    msgs = sim_run_engine.simulate_plan(i04_grid_detect_then_xrc_default_params())

    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "stage"
        and msg.obj.name == "zocalo"
        and msg.kwargs["group"] == ZOCALO_STAGE_GROUP,
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "unstage" and msg.obj.name == "zocalo",
    )


@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.ispyb_activation_wrapper",
    lambda plan, _: plan,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
    lambda *_: fake_generator(),
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
    lambda *_: fake_generator(),
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_results_and_move_to_xtal"
)
def test_detect_grid_and_do_gridscan_gives_params_specified_grid(
    mock_change_aperture_then_move_to_xtal: MagicMock,
    mock_create_flyscan_params: MagicMock,
    test_fgs_params: SpecifiedThreeDGridScan,
    run_engine: RunEngine,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
):
    mock_create_flyscan_params.return_value = test_fgs_params
    run_engine(
        i04_grid_detect_then_xrc_default_params(
            udc=False,
        )
    )
    mock_change_aperture_then_move_to_xtal.assert_called_once()
    assert mock_change_aperture_then_move_to_xtal.call_args[0][1] == test_fgs_params


@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.ispyb_activation_wrapper",
    lambda plan, _: plan,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
    lambda *_: fake_generator(),
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
    lambda *_: fake_generator(),
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
)
@patch(
    "mx_bluesky.common.experiment_plans.change_aperture_then_move_plan._get_xrc_results",
)
def test_i04_grid_detect_then_xrc_only_sets_aperture_at_start_of_plan(
    mock_get_xrc_results: MagicMock,
    mock_create_flyscan_params: MagicMock,
    test_fgs_params: SpecifiedThreeDGridScan,
    run_engine: RunEngine,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
):
    aperture_scatterguard: ApertureScatterguard = (
        i04_grid_detect_then_xrc_default_params.keywords["aperture_scatterguard"]
    )
    get_mock_put(aperture_scatterguard.selected_aperture).reset_mock()

    def fake_get_xrc_results_plan():
        results = [
            XRayCentreResult(
                centre_of_mass_mm=np.array([0.05, 0.15, 0.25]),
                bounding_box_mm=(
                    np.array([0.15, 0.15, 0.15]),
                    np.array([0.75, 0.75, 0.65]),
                ),
                max_count=105062,
                total_count=2387574,
                sample_id=12345,
            )
        ]
        yield from ()
        return results

    mock_get_xrc_results.return_value = fake_get_xrc_results_plan()

    mock_create_flyscan_params.return_value = test_fgs_params
    run_engine(
        i04_grid_detect_then_xrc_default_params(
            udc=False,
        )
    )

    # Out of beam for grid detect, large for gridscan, out again after tidy up
    assert get_mock_put(aperture_scatterguard.selected_aperture).call_args_list == [
        call(ApertureValue.OUT_OF_BEAM),
        call(ApertureValue.LARGE),
        call(ApertureValue.OUT_OF_BEAM),
    ]



"""
This module contains the parameter models exported via the hyperion-blueapi REST interface.
"""

from pydantic import BaseModel
from pydantic.config import ConfigDict

from mx_bluesky.common.parameters.components import (
    get_param_version,
)
from mx_bluesky.hyperion.blueapi.mixins import WithCentreSelection
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect


class HyperionParam(BaseModel):
    model_config = ConfigDict(extra="forbid")


class RobotLoadThenCentreParams(HyperionParam):
    storage_directory: str
    file_name: str
    transmission_frac: float
    exposure_time_s: float
    omega_start_deg: float
    chi_start_deg: float
    tip_offset_um: float
    grid_width_um: float


class SingleRotationScanParams(HyperionParam):
    omega_start_deg: float
    phi_start_deg: float
    chi_start_deg: float
    rotation_direction: str
    scan_width_deg: float


class MultiRotationScanParams(HyperionParam):
    comment: str
    file_name: str
    storage_directory: str
    exposure_time_s: float
    rotation_increment_deg: float
    snapshot_omegas_deg: list[float]
    rotation_scans: list[SingleRotationScanParams]
    transmission_frac: float
    ispyb_experiment_type: str


class LoadCentreCollectParams(WithCentreSelection, HyperionParam):
    """This model is exposed as the BlueAPI REST parameter model for Hyperion Collections.
    It can represent the full range of operations that are supported by LoadCentreCollect;
    this is a superset of the operations that are actually required to follow Agamemnon instructions.

    This is intended to provide additional flexibility to allow a potential future point of configuration
    where the supervisor can supply default values for values not explicitly specified by Agamemnon,
    also to allow future exposure of this BlueAPI plan for e.g. commissioning purposes.
    This may also permit the supervisor to implement future functionality by adjusting these parameters.
    """

    visit: str
    detector_distance_mm: float
    sample_id: int
    sample_puck: int
    sample_pin: int
    demand_energy_ev: float
    robot_load_then_centre: RobotLoadThenCentreParams
    multi_rotation_scan: MultiRotationScanParams


def load_centre_collect_to_internal(
    external_params: LoadCentreCollectParams,
) -> LoadCentreCollect:
    params_as_dict = external_params.model_dump()
    params_as_dict["parameter_model_version"] = get_param_version()
    return LoadCentreCollect(**params_as_dict)



from daq_config_server.models.feature_settings.i04_feature_settings import (
    I04FeatureSettings,
)
from dodal.beamlines.i04 import DAQ_CONFIGURATION_PATH
from dodal.common.beamlines.config_client import get_config_client

GDA_DOMAIN_PROPERTIES_PATH = DAQ_CONFIGURATION_PATH + "/domain/domain.properties"


def get_i04_feature_settings() -> I04FeatureSettings:
    config_client = get_config_client("i04")
    return config_client.get_file_contents(
        GDA_DOMAIN_PROPERTIES_PATH,
        desired_return_type=I04FeatureSettings,
        reset_cached_result=True,
    )



import json
from pathlib import Path
from unittest.mock import patch

import pytest
from dodal.devices.aperturescatterguard import ApertureValue
from pydantic import ValidationError

from mx_bluesky.common.external_interaction.callbacks.common.grid_detection_callback import (
    GridParamUpdate,
)
from mx_bluesky.common.parameters.constants import GridscanParamConstants
from mx_bluesky.common.parameters.rotation import (
    SingleRotationScan,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    HyperionSpecifiedThreeDGridScan,
    OddYStepsError,
)
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadThenCentre

from ....conftest import raw_params_from_file


@pytest.fixture(autouse=True)
def always_use_beamline_i03(use_beamline_i03): ...


@pytest.fixture
def load_centre_collect_params_with_panda(tmp_path, request):
    with patch(
        "mx_bluesky.common.parameters.constants.GDA_DOMAIN_PROPERTIES_PATH",
        new="tests/test_data/test_domain_properties_with_panda",
    ):
        params = raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_load_centre_collect_params.json",
            tmp_path,
        )
        params["features"]["use_panda_for_gridscan"] = True
        if params_dict := getattr(request, "param", {}):
            for k, v in params_dict.items():
                params.setdefault("features", {})[k] = v
        return LoadCentreCollect(**params)


@pytest.fixture()
def minimal_3d_gridscan_params():
    return {
        "sample_id": 123,
        "x_start_um": 0.123,
        "y_start_um": 0.777,
        "z_start_um": 0.05,
        "parameter_model_version": "5.0.0",
        "visit": "cm12345",
        "file_name": "test_file_name",
        "y2_start_um": 2,
        "z2_start_um": 2,
        "x_steps": 5,
        "y_steps": 7,
        "z_steps": 9,
        "storage_directory": "/tmp/dls/i03/data/2024/cm31105-4/xraycentring/123456/",
    }


def get_empty_grid_parameters() -> GridParamUpdate:
    return {
        "x_start_um": 1,
        "y_start_um": 1,
        "y2_start_um": 1,
        "z_start_um": 1,
        "z2_start_um": 1,
        "x_steps": 1,
        "y_steps": 1,
        "z_steps": 1,
        "x_step_size_um": 1,
        "y_step_size_um": 1,
        "z_step_size_um": 1,
    }


def test_minimal_3d_gridscan_params(minimal_3d_gridscan_params):
    test_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    assert {"sam_x", "sam_y", "sam_z"} == set(test_params.scan_points.keys())
    assert test_params.scan_indices == [0, 35]
    assert test_params.num_images == (5 * 7 + 5 * 9)
    assert test_params.exposure_time_s == GridscanParamConstants.EXPOSURE_TIME_S


def test_cant_do_panda_fgs_with_odd_y_steps(minimal_3d_gridscan_params):
    test_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    with pytest.raises(OddYStepsError):
        _ = test_params.panda_fast_gridscan_params
    assert test_params.fast_gridscan_params


def test_serialise_deserialise(minimal_3d_gridscan_params):
    test_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    serialised = json.loads(test_params.model_dump_json())
    deserialised = HyperionSpecifiedThreeDGridScan(**serialised)
    assert deserialised.demand_energy_ev is None
    assert deserialised.visit == "cm12345"
    assert deserialised.x_start_um == 0.123


@pytest.mark.parametrize(
    "version, valid",
    [
        ("4.3.0", False),
        ("6.3.7", False),
        ("5.0.0", True),
        ("5.3.0", True),
        ("5.3.7", True),
    ],
)
def test_param_version(minimal_3d_gridscan_params, version: str, valid: bool):
    minimal_3d_gridscan_params["parameter_model_version"] = version
    if valid:
        _ = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    else:
        with pytest.raises(ValidationError):
            _ = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)


def test_robot_load_then_centre_params():
    params = {
        "parameter_model_version": "5.0.0",
        "sample_id": 123456,
        "visit": "cm12345",
        "file_name": "file_name",
        "storage_directory": "/tmp/dls/i03/data/2024/cm31105-4/xraycentring/123456/",
    }
    params["detector_distance_mm"] = 200
    test_params = RobotLoadThenCentre(**params)  # type:ignore
    assert test_params.detector_params


def test_default_snapshot_path(minimal_3d_gridscan_params):
    gridscan_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    assert gridscan_params.snapshot_directory == Path(
        "/tmp/dls/i03/data/2024/cm31105-4/xraycentring/123456/snapshots"
    )

    params_with_snapshot_path = dict(minimal_3d_gridscan_params)
    params_with_snapshot_path["snapshot_directory"] = "/tmp/my_snapshots"

    gridscan_params_with_snapshot_path = HyperionSpecifiedThreeDGridScan(
        **params_with_snapshot_path
    )
    assert gridscan_params_with_snapshot_path.snapshot_directory == Path(
        "/tmp/my_snapshots"
    )


def test_osc_is_used(tmp_path):
    raw_params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json",
        tmp_path,
    )
    for osc in [0.001, 0.05, 0.1, 0.2, 0.75, 1, 1.43]:
        raw_params["rotation_increment_deg"] = osc
        params = SingleRotationScan(**raw_params)
        assert params.rotation_increment_deg == osc
        assert params.num_images == int(params.scan_width_deg / osc)


def test_selected_aperture_uses_default(tmp_path):
    raw_params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json",
        tmp_path,
    )
    raw_params["selected_aperture"] = None
    params = SingleRotationScan(**raw_params)
    assert params.selected_aperture == ApertureValue.LARGE


@pytest.mark.parametrize(
    "enable_gpu",
    [
        True,
        False,
    ],
)
@patch("mx_bluesky.common.parameters.components.os")
def test_dev_shm_enabled_if_use_gpu_results_enabled(
    _, enable_gpu, minimal_3d_gridscan_params
):
    minimal_3d_gridscan_params["detector_distance_mm"] = 100
    properties_path = (
        "tests/test_data/test_domain_properties_with_no_gpu"
        if not enable_gpu
        else "tests/test_data/test_domain_properties"
    )
    with patch(
        "mx_bluesky.hyperion.external_interaction.config_server.GDA_DOMAIN_PROPERTIES_PATH",
        new=properties_path,
    ):
        grid_scan = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
        assert grid_scan.detector_params.enable_dev_shm == enable_gpu



from __future__ import annotations

import os
import re
import signal
import subprocess
import sys
import threading
from collections.abc import Mapping
from genericpath import isfile
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from time import sleep
from typing import Any
from unittest.mock import MagicMock, patch

import bluesky.plan_stubs as bps
import pytest
import zmq
from blueapi.config import ApplicationConfig, ConfigLoader
from blueapi.core import BlueskyContext, DataEvent, EventPublisher
from bluesky import preprocessors as bpp
from bluesky.callbacks import CallbackBase
from bluesky.callbacks.zmq import Publisher
from bluesky.run_engine import RunEngine
from bluesky.utils import MsgGenerator
from bluesky_stomp.messaging import MessageContext, StompClient
from bluesky_stomp.models import Broker, MessageTopic
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.smargon import Smargon
from zmq.utils.monitor import recv_monitor_message

from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    common_flyscan_xray_centre,
)
from mx_bluesky.common.experiment_plans.inner_plans.xrc_results_utils import (
    fetch_xrc_results_from_zocalo,
    zocalo_stage_decorator,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    ispyb_activation_decorator,
)
from mx_bluesky.common.parameters.components import WithSnapshot
from mx_bluesky.common.parameters.rotation import (
    RotationScan,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.utils import convert_angstrom_to_ev
from mx_bluesky.hyperion.experiment_plans.hyperion_flyscan_xray_centre_plan import (
    construct_hyperion_specific_features,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
    rotation_scan_internal,
)
from mx_bluesky.hyperion.external_interaction.callbacks.stomp.dispatcher import (
    BLUEAPI_EVENT_TOPIC,
)
from mx_bluesky.hyperion.parameters.constants import CONST, HyperionConstants
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

from .....conftest import fake_read
from ..conftest import fetch_comment  # noqa  # type: ignore
from .stomp_callback_wrapper.__main__ import CALLBACK_TOPIC

"""
Note that because these tests use the external processes some of the errors coming from
them may not be very informative. You will want to check the log files produced in `tmp`
for better logs.
"""


TASK_ID = "TASK_ID"


class DocumentCatcher(CallbackBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start = MagicMock()
        self.descriptor = MagicMock()
        self.event = MagicMock()
        self.stop = MagicMock()


def rotation_scan(
    composite: RotationScanComposite,
    parameters: RotationScan,
    oav_params: OAVParameters | None = None,
) -> MsgGenerator:
    @bpp.set_run_key_decorator(CONST.PLAN.ROTATION_MULTI_OUTER)
    @bpp.run_decorator(
        md={
            "activate_callbacks": ["BeamDrawingCallback"],
            "with_snapshot": parameters.model_dump_json(
                include=WithSnapshot.model_fields.keys()  # type: ignore
            ),
        }
    )
    def _wrapped_rotation_scan():
        yield from rotation_scan_internal(composite, parameters, oav_params)

    yield from _wrapped_rotation_scan()


def event_monitor(monitor: zmq.Socket, connection_active_lock: threading.Lock) -> None:
    try:
        while monitor.poll():
            monitor_event = recv_monitor_message(monitor)
            LOGGER.info(f"Event: {monitor_event}")
            if monitor_event["event"] == zmq.EVENT_CONNECTED:
                LOGGER.info("CONNECTED - acquiring connection_active_lock")
                connection_active_lock.acquire()
            if monitor_event["event"] == zmq.EVENT_MONITOR_STOPPED:
                break
    except zmq.ZMQError:
        pass
    finally:
        connection_active_lock.release()
        monitor.close()
        LOGGER.info("event monitor thread done!")


def external_callbacks_with_args(*args):
    process_env = os.environ.copy()
    external_callbacks_process = subprocess.Popen(
        [sys.executable, *args], env=process_env
    )
    yield external_callbacks_process
    external_callbacks_process.send_signal(signal.SIGINT)
    sleep(0.01)
    external_callbacks_process.kill()
    external_callbacks_process.wait(10)


@pytest.fixture
def external_callbacks():
    yield from external_callbacks_with_args(
        "-m",
        "mx_bluesky.hyperion.external_interaction.callbacks",
        "--dev",
        "--watchdog-port",
        str(HyperionConstants.HYPERION_PORT),
    )


@pytest.fixture
def external_callbacks_with_stomp():
    yield from external_callbacks_with_args(
        "tests/system_tests/hyperion/external_interaction/callbacks/stomp_callback_wrapper/__main__.py",
        "--dev",
        "--watchdog-port",
        str(HyperionConstants.HYPERION_PORT),
        "--stomp-config",
        "tests/test_data/stomp_callback_test_config.yaml",
    )


@pytest.fixture
def bluesky_context_with_stomp(run_engine: RunEngine):
    loader = ConfigLoader(ApplicationConfig)
    loader.use_values_from_yaml(Path("tests/test_data/stomp_callback_test_config.yaml"))
    config = loader.load()
    context = BlueskyContext(configuration=config, run_engine=run_engine)
    assert config.stomp.url.host is not None and config.stomp.url.port is not None
    stomp_client = StompClient.for_broker(
        broker=Broker(
            host=config.stomp.url.host,
            port=config.stomp.url.port,
            auth=config.stomp.auth,
        )
    )

    event_publisher = EventPublisher()

    def fire_event(name: str, doc: Mapping[str, Any]):
        event_publisher.publish(DataEvent(name=name, doc=doc, task_id=TASK_ID))

    re_sub_id = context.run_engine.subscribe(fire_event)
    stomp_client.connect()

    def send_message(event, correlation_id):
        stomp_client.send(MessageTopic(name=BLUEAPI_EVENT_TOPIC), event, correlation_id)

    publisher_sub_id = event_publisher.subscribe(send_message)

    try:
        yield context, stomp_client
    finally:
        event_publisher.unsubscribe(publisher_sub_id)
        stomp_client.disconnect()
        context.run_engine.unsubscribe(re_sub_id)


@pytest.fixture
def run_engine_with_external_callbacks(
    external_callbacks,
    zocalo_env,  # ZOCALO_CONFIG must be exported to external callback environment
    run_engine,
):
    publisher = Publisher(f"localhost:{CONST.CALLBACK_0MQ_PROXY_PORTS[0]}")
    monitor = publisher._socket.get_monitor_socket()

    connection_active_lock = threading.Lock()
    t = threading.Thread(
        target=event_monitor,
        args=(monitor, connection_active_lock),
        name="event_monitor",
    )
    t.start()

    while not connection_active_lock.locked():
        sleep(0.1)  # wait for connection to happen before continuing

    sub_id = run_engine.subscribe(publisher)

    yield run_engine

    run_engine.unsubscribe(sub_id)
    publisher.close()

    t.join()


@pytest.mark.system_test
def test_run_engine_with_external_callbacks_starts_and_stops(
    run_engine_with_external_callbacks: RunEngine,
):
    run_engine = run_engine_with_external_callbacks

    def plan():
        yield from bps.sleep(1)

    run_engine(plan())


@pytest.mark.system_test
async def test_external_callbacks_handle_gridscan_ispyb_and_zocalo(
    oav_for_system_test: OAV,
    smargon: Smargon,
    run_engine_with_external_callbacks: RunEngine,
    dummy_params: HyperionSpecifiedThreeDGridScan,
    fgs_composite_for_fake_zocalo: HyperionFlyScanXRayCentreComposite,
    fetch_comment,  # noqa
    fetch_datacollection_ids_for_group_id,
    fake_grid_snapshot_plan,
):
    """
    This test requires fake zocalo, and a connection to the dev ISPyB database.
    """

    run_engine = run_engine_with_external_callbacks

    doc_catcher = DocumentCatcher()
    run_engine.subscribe(doc_catcher)

    # Run the xray centring plan
    beamline_specific = construct_hyperion_specific_features(
        fgs_composite_for_fake_zocalo, dummy_params
    )

    @zocalo_stage_decorator(fgs_composite_for_fake_zocalo.zocalo)
    @ispyb_activation_decorator(dummy_params)
    def wrapped_xray_centre():
        yield from fake_grid_snapshot_plan(smargon, oav_for_system_test)
        yield from common_flyscan_xray_centre(
            fgs_composite_for_fake_zocalo, dummy_params, beamline_specific
        )
        yield from fetch_xrc_results_from_zocalo(
            fgs_composite_for_fake_zocalo.zocalo, dummy_params
        )

    run_engine(wrapped_xray_centre())

    # get dcids from zocalo device
    dcid_reading = await fgs_composite_for_fake_zocalo.zocalo.ispyb_dcid.read()
    dcgid_reading = await fgs_composite_for_fake_zocalo.zocalo.ispyb_dcgid.read()

    dcid = dcid_reading["zocalo-ispyb_dcid"]["value"]
    dcgid = dcgid_reading["zocalo-ispyb_dcgid"]["value"]

    assert dcid != 0
    assert dcgid != 0

    dcids = fetch_datacollection_ids_for_group_id(dcgid)

    assert dcid in dcids

    # check the data in dev ispyb corresponding to this "collection"
    ispyb_comment1 = fetch_comment(dcids[0])
    ispyb_comment2 = fetch_comment(dcids[1])
    assert ispyb_comment1 != "" and ispyb_comment2 != ""
    assert (
        "Zocalo processing took" in ispyb_comment1
        or "Zocalo processing took" in ispyb_comment2
    )


@pytest.mark.system_test
def test_remote_callbacks_write_to_dev_ispyb_for_rotation(
    run_engine_with_external_callbacks: RunEngine,
    params_for_rotation_scan: RotationScan,
    fetch_comment,  # noqa
    fetch_datacollection_attribute,
    composite_for_rotation_scan,
    oav_parameters_for_rotation,
):
    test_wl = 0.71
    test_bs_x = 0.020
    test_bs_y = 0.020
    test_exp_time = 0.023
    test_img_wid = 0.27

    params_for_rotation_scan.rotation_increment_deg = test_img_wid
    params_for_rotation_scan.exposure_time_s = test_exp_time
    params_for_rotation_scan.demand_energy_ev = convert_angstrom_to_ev(test_wl)

    with patch("bluesky.preprocessors.__read_and_stash_a_motor", fake_read):
        run_engine_with_external_callbacks(
            rotation_scan(
                composite_for_rotation_scan,
                params_for_rotation_scan,
                oav_parameters_for_rotation,
            )
        )

    sleep(1)
    assert isfile("/tmp/logs/bluesky/hyperion_ispyb_callback.log")
    ispyb_log_tail = subprocess.run(
        ["tail", "/tmp/logs/bluesky/hyperion_ispyb_callback.log"],
        timeout=1,
        stdout=subprocess.PIPE,
    ).stdout.decode("utf-8")

    ids_re = re.compile(
        r"data_collection_ids=\((\d+),\) data_collection_group_id=(\d+) "
    )
    matches = ids_re.findall(ispyb_log_tail)

    dcid = matches[0][0]

    comment = fetch_comment(dcid)
    assert comment == "test Sample position (µm): (1, 2, 3) Aperture: Small. "
    wavelength = fetch_datacollection_attribute(dcid, "wavelength")
    beamsize_x = fetch_datacollection_attribute(dcid, "beamSizeAtSampleX")
    beamsize_y = fetch_datacollection_attribute(dcid, "beamSizeAtSampleY")
    exposure = fetch_datacollection_attribute(dcid, "exposureTime")

    assert wavelength == test_wl
    assert beamsize_x == test_bs_x
    assert beamsize_y == test_bs_y
    assert exposure == test_exp_time


@pytest.mark.system_test
def test_external_callbacks_ping(external_callbacks):
    ping_received = threading.Event()

    class PingHTTPRequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):  # noqa: N802
            assert self.path == "/callbackPing"
            ping_received.set()

    httpd = HTTPServer(("localhost", CONST.HYPERION_PORT), PingHTTPRequestHandler)
    server_thread = threading.Thread(
        group=None, target=httpd.serve_forever, daemon=True
    )
    try:
        server_thread.start()
        ping_received.wait(5)
    finally:
        httpd.shutdown()
        server_thread.join()


@pytest.mark.system_test
def test_external_callbacks_with_stomp(
    external_callbacks_with_stomp, bluesky_context_with_stomp
):
    bluesky_context, stomp_client = bluesky_context_with_stomp

    def read_the_baton():
        yield from bps.open_run(md={"run_name": "READ BATON PLAN"})
        baton = bluesky_context.find_device("baton")
        yield from bps.abs_set(baton.requested_user, "TEST_USER")
        yield from bps.create(name="BATON_READ_EVENT")
        yield from bps.read(baton)
        yield from bps.save()
        yield from bps.close_run()

    msgs = []

    started = threading.Event()
    stopped = threading.Event()

    def on_msg_from_callback(msg: str, context: MessageContext):
        msgs.append(msg)
        if msg == "stop":
            stopped.set()

    sub_id = stomp_client.subscribe(
        MessageTopic(name=CALLBACK_TOPIC), on_msg_from_callback
    )

    try:
        started.wait(10)
        bluesky_context.run_engine(read_the_baton())
        stopped.wait(10)
    finally:
        stomp_client.unsubscribe(sub_id)

    assert msgs == ["init", "start: READ BATON PLAN", "event: TEST_USER", "stop"]



import pytest
from bluesky import RunEngine
from bluesky.plan_stubs import null

from mx_bluesky.common.utils.exceptions import (
    WarningError,
    catch_exception_and_warn,
)


class _TestError(Exception):
    pass


def dummy_plan():
    yield from null()
    raise _TestError


def test_catch_exception_and_warn_correctly_raises_warning_exception(
    run_engine: RunEngine,
):
    with pytest.raises(WarningError):
        run_engine(catch_exception_and_warn(_TestError, dummy_plan))


def test_catch_exception_and_warn_correctly_raises_original_exception(
    run_engine: RunEngine,
):
    with pytest.raises(_TestError):
        run_engine(catch_exception_and_warn(ValueError, dummy_plan))



import asyncio
from unittest.mock import MagicMock

import bluesky.preprocessors as bpp
import pytest
from bluesky import RunEngine
from bluesky import plan_stubs as bps
from bluesky.utils import FailedStatus
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.xbpm_feedback import Pause, XBPMFeedback
from dodal.plans.preprocessors.verify_undulator_gap import (
    verify_undulator_gap_before_run_decorator,
)
from ophyd_async.core import AsyncStatus, completed_status, set_mock_value

from mx_bluesky.common.device_setup_plans.xbpm_feedback import (
    unpause_xbpm_feedback_and_set_transmission_to_1,
)
from mx_bluesky.common.preprocessors.preprocessors import (
    pause_xbpm_feedback_during_collection_at_desired_transmission_decorator,
)
from tests.conftest import XBPMAndTransmissionWrapperComposite


@pytest.fixture
def composite(
    xbpm_and_transmission_wrapper_composite: XBPMAndTransmissionWrapperComposite,
) -> XBPMAndTransmissionWrapperComposite:
    xbpm_and_transmission_wrapper_composite.undulator.set = MagicMock(
        side_effect=lambda _: completed_status()
    )

    return xbpm_and_transmission_wrapper_composite


async def test_xbpm_decorator_with_undulator_check_decorators(
    run_engine, composite: XBPMAndTransmissionWrapperComposite
):
    energy_in_kev = 11.3
    composite.dcm.energy_in_keV.user_readback.read = MagicMock(
        return_value={"value": {"value": energy_in_kev}}
    )

    @pause_xbpm_feedback_during_collection_at_desired_transmission_decorator(
        composite, 0.1
    )
    @verify_undulator_gap_before_run_decorator(composite)
    @bpp.run_decorator()
    def my_collection_plan():
        yield from bps.null()

    set_mock_value(composite.xbpm_feedback.pos_stable, 1)
    run_engine(my_collection_plan())

    # Stop pyright from complaining
    assert isinstance(composite.xbpm_feedback.trigger, MagicMock)
    assert isinstance(composite.undulator.set, MagicMock)

    # Assert XBPM is stable
    composite.xbpm_feedback.trigger.assert_called_once()
    # Assert DCM energy is read after XBPM is stable
    composite.dcm.energy_in_keV.user_readback.read.assert_called_once()
    # Assert Undulator is finally set
    composite.undulator.set.assert_called_once()
    # Assert energy passed to the Undulator is the same as read from the DCM
    assert composite.undulator.set.call_args.args[0] == energy_in_kev


async def test_given_xpbm_checks_pass_when_plan_run_with_decorator_then_run_as_expected(
    run_engine, composite: XBPMAndTransmissionWrapperComposite
):
    expected_transmission = 0.3

    @pause_xbpm_feedback_during_collection_at_desired_transmission_decorator(
        composite, expected_transmission
    )
    @bpp.run_decorator()
    def my_collection_plan():
        read_transmission = yield from bps.rd(composite.attenuator.actual_transmission)
        assert read_transmission == expected_transmission
        pause_feedback = yield from bps.rd(composite.xbpm_feedback.pause_feedback)
        assert pause_feedback == Pause.PAUSE

    set_mock_value(composite.xbpm_feedback.pos_stable, 1)

    run_engine(my_collection_plan())

    assert await composite.attenuator.actual_transmission.get_value() == 1.0
    assert await composite.xbpm_feedback.pause_feedback.get_value() == Pause.RUN


async def test_given_xbpm_checks_fail_when_plan_run_with_decorator_then_plan_not_run(
    run_engine, composite: XBPMAndTransmissionWrapperComposite
):
    mock = MagicMock()

    @pause_xbpm_feedback_during_collection_at_desired_transmission_decorator(
        composite, 0.1
    )
    @bpp.run_decorator()
    def my_collection_plan():
        mock()
        yield from bps.null()

    composite.xbpm_feedback.trigger = MagicMock(
        side_effect=lambda: completed_status(Exception())
    )

    with pytest.raises(FailedStatus):
        run_engine(my_collection_plan())

    mock.assert_not_called()
    assert await composite.attenuator.actual_transmission.get_value() == 1.0
    assert await composite.xbpm_feedback.pause_feedback.get_value() == Pause.RUN


async def test_given_xpbm_checks_pass_and_plan_fails_when_plan_run_with_decorator_then_cleaned_up(
    run_engine, composite: XBPMAndTransmissionWrapperComposite
):
    set_mock_value(composite.xbpm_feedback.pos_stable, 1)

    class MyError(Exception):
        pass

    @pause_xbpm_feedback_during_collection_at_desired_transmission_decorator(
        composite, 0.1
    )
    @bpp.run_decorator()
    def my_collection_plan():
        yield from bps.null()
        raise MyError()

    with pytest.raises(MyError):
        run_engine(my_collection_plan())

    assert await composite.attenuator.actual_transmission.get_value() == 1.0
    assert await composite.xbpm_feedback.pause_feedback.get_value() == Pause.RUN


def test_unpause_feedback_and_set_transmission_to_1_times_out_if_timeout_specified(
    run_engine: RunEngine,
    xbpm_feedback: XBPMFeedback,
    attenuator: BinaryFilterAttenuator,
):
    @AsyncStatus.wrap
    async def wait_for_1_s():
        await asyncio.sleep(1)

    xbpm_feedback.trigger = MagicMock(side_effect=wait_for_1_s)
    with pytest.raises(TimeoutError):
        run_engine(
            unpause_xbpm_feedback_and_set_transmission_to_1(
                xbpm_feedback=xbpm_feedback,
                attenuator=attenuator,
                timeout_for_stable=0.1,
            )
        )



from collections.abc import Callable, Sequence
from functools import partial
from typing import Any

from blueapi.core import BlueskyContext
from bluesky import plan_stubs as bps
from bluesky.utils import MsgGenerator, RequestAbort
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.motors import XYZStage
from dodal.devices.robot import BartRobot
from dodal.devices.smargon import Smargon
from pydantic import BaseModel

from mx_bluesky.common.parameters.constants import Status
from mx_bluesky.common.utils.context import (
    device_composite_from_context,
    find_device_in_context,
)
from mx_bluesky.common.utils.exceptions import WarningError
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion._plan_runner_params import (
    RobotUnload,
    UDCCleanup,
    UDCDefaultState,
    Wait,
)
from mx_bluesky.hyperion.blueapi.in_process import (
    LoadCentreCollectParams,
    clean_up_udc,
    load_centre_collect,
    move_to_udc_default_state,
    robot_unload,
)
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    create_devices,
)
from mx_bluesky.hyperion.experiment_plans.udc_default_state import UDCDefaultDevices
from mx_bluesky.hyperion.plan_runner import PlanError, PlanRunner


class InProcessRunner(PlanRunner):
    """Runner that executes experiments from inside a running Bluesky plan"""

    def __init__(self, context: BlueskyContext, dev_mode: bool) -> None:
        super().__init__(context, dev_mode)
        self._current_status: Status = Status.IDLE

    def decode_and_execute(
        self, current_visit: str | None, parameter_list: Sequence[BaseModel]
    ) -> MsgGenerator:
        current_visit = current_visit or ""
        try:
            for parameters in parameter_list:
                LOGGER.info(
                    f"Executing plan with parameters: {parameters.model_dump_json(indent=2)}"
                )
                match parameters:
                    case LoadCentreCollectParams():
                        current_visit = parameters.visit
                        devices: Any = create_devices(self.context)
                        yield from self.execute_plan(
                            partial(
                                load_centre_collect,
                                parameters,
                                devices,
                            )
                        )
                    case RobotUnload():
                        yield from _robot_unload(self.context, current_visit)
                    case Wait():
                        yield from self.execute_plan(partial(_runner_sleep, parameters))
                    case UDCDefaultState():
                        udc_default_devices: UDCDefaultDevices = (
                            device_composite_from_context(
                                self.context, UDCDefaultDevices
                            )
                        )
                        yield from move_to_udc_default_state(udc_default_devices)
                    case UDCCleanup():
                        yield from _clean_up_udc(self.context)
                    case _:
                        raise AssertionError(
                            f"Unsupported instruction decoded from agamemnon {type(parameters)}"
                        )
        except WarningError:
            pass
        return current_visit

    def execute_plan(
        self,
        experiment: Callable[[], MsgGenerator],
    ) -> MsgGenerator:
        """Execute the specified experiment plan.
        Args:
            experiment: The experiment to run
        Raises:
            PlanError: If the plan raised an exception
            RequestAbort: If the RunEngine aborted during execution"""

        self._current_status = Status.BUSY

        try:
            yield from self.check_external_callbacks_are_alive()
            yield from experiment()
            self._current_status = Status.IDLE
        except WarningError as e:
            LOGGER.warning("Plan failed with warning", exc_info=e)
            self._current_status = Status.FAILED
            raise
        except RequestAbort:
            # This will occur when the run engine processes an abort when we shut down
            LOGGER.info("UDC Runner aborting")
            raise
        except Exception as e:
            LOGGER.error("Plan failed with exception", exc_info=e)
            self._current_status = Status.FAILED
            raise PlanError("Exception thrown in plan execution") from e

    def shutdown(self):
        """Performs a prompt shutdown. Aborts the run engine and terminates the loop
        waiting for messages."""

        LOGGER.info("Shutting down: Stopping the run engine gracefully")
        if self.current_status != Status.ABORTING:
            self._current_status = Status.ABORTING
            self.request_run_engine_abort()
            return

    @property
    def current_status(self) -> Status:
        return self._current_status


def _runner_sleep(parameters: Wait) -> MsgGenerator:
    yield from bps.sleep(parameters.duration_s)


def _robot_unload(context: BlueskyContext, visit: str) -> MsgGenerator:
    robot = find_device_in_context(context, "robot", BartRobot)
    smargon = find_device_in_context(context, "gonio", Smargon)
    aperture_scatterguard = find_device_in_context(
        context, "aperture_scatterguard", ApertureScatterguard
    )
    lower_gonio = find_device_in_context(context, "lower_gonio", XYZStage)
    yield from robot_unload(
        visit,
        robot=robot,
        smargon=smargon,
        aperture_scatterguard=aperture_scatterguard,
        lower_gonio=lower_gonio,
    )


def _clean_up_udc(context: BlueskyContext) -> MsgGenerator:
    detector_motion = find_device_in_context(context, "detector_motion", DetectorMotion)
    yield from clean_up_udc(
        detector_motion=detector_motion,
    )



import json
from collections.abc import Generator
from math import isclose
from pathlib import PosixPath
from typing import cast
from unittest.mock import MagicMock, patch

import pytest
from dodal.devices.zebra.zebra import RotationDirection

from mx_bluesky.common.parameters.constants import GridscanParamConstants
from mx_bluesky.hyperion._plan_runner_params import Wait
from mx_bluesky.hyperion.blueapi.parameters import (
    LoadCentreCollectParams,
    load_centre_collect_to_internal,
)
from mx_bluesky.hyperion.external_interaction.agamemnon import (
    _get_next_instruction,
    _get_pin_type_from_agamemnon_collect_parameters,
    _get_withenergy_parameters_from_agamemnon,
    _get_withvisit_parameters_from_agamemnon,
    _instruction_and_data,
    _PinType,
    _SinglePin,
    compare_params,
    create_parameters_from_agamemnon,
    update_params_from_agamemnon,
)
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect


@pytest.mark.parametrize(
    "num_wells, well_width, buffer, expected_width",
    [
        (3, 500, 0, 1000),
        (6, 50, 100, 450),
        (2, 800, 50, 900),
    ],
)
def test_given_various_pin_formats_then_pin_width_as_expected(
    num_wells, well_width, buffer, expected_width
):
    pin = _PinType(num_wells, well_width, buffer)
    assert pin.full_width == expected_width


def set_up_agamemnon_params(
    loop_type: str | None = None,
    prefix: str | None = None,
    distance: int | None = None,
    wavelength: float | None = None,
):
    return {
        "collection": [{"distance": distance, "wavelength": wavelength}],
        "prefix": prefix,
        "sample": {"loopType": loop_type, "id": 1, "position": 1, "container": 1},
    }


def test_given_no_loop_type_in_parameters_then_single_pin_returned():
    assert (
        _get_pin_type_from_agamemnon_collect_parameters(set_up_agamemnon_params())
        == _SinglePin()
    )


@pytest.mark.parametrize(
    "loop_name, expected_loop",
    [
        ("multipin_6x50+9", _PinType(6, 50, 9)),
        ("multipin_6x25.8+8.6", _PinType(6, 25.8, 8.6)),
        ("multipin_9x31+90", _PinType(9, 31, 90)),
    ],
)
def test_given_multipin_loop_type_in_parameters_then_expected_pin_returned(
    loop_name: str, expected_loop: _PinType
):
    assert (
        _get_pin_type_from_agamemnon_collect_parameters(
            set_up_agamemnon_params(loop_name)
        )
        == expected_loop
    )


@pytest.mark.parametrize(
    "loop_name",
    [
        "nonesense",
        "single_pin_78x89+1",
    ],
)
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
def test_given_completely_unrecognised_loop_type_in_parameters_then_warning_logged_single_pin_returned(
    mock_logger: MagicMock,
    loop_name: str,
):
    assert (
        _get_pin_type_from_agamemnon_collect_parameters(
            set_up_agamemnon_params(loop_name)
        )
        == _SinglePin()
    )
    mock_logger.warning.assert_called_once()


@pytest.mark.parametrize(
    "loop_name",
    [
        "multipin_67x56",
        "multipin_90+4",
        "multipin_8",
        "multipin_6x50+",
        "multipin_6x50+98.",
        "multipin_6x50+.1",
        "multipin_6x.50+98",
        "multipin_6x50+98.1.2",
        "multipin_6x50.5.6+98",
        "multipin_6x50+98..1",
        "multipin_6x.50+.98",
        "multipin_6x+98",
    ],
)
def test_given_unrecognised_multipin_in_parameters_then_warning_logged_single_pin_returned(
    loop_name: str,
):
    with pytest.raises(ValueError) as e:
        _get_pin_type_from_agamemnon_collect_parameters(
            set_up_agamemnon_params(loop_name)
        )
    assert "Expected multipin format" in str(e.value)


def configure_mock_agamemnon(mock_requests: MagicMock, loop_type: str | None):
    mock_requests.get.return_value.content = json.dumps(
        {"collect": set_up_agamemnon_params(loop_type, "", 255, 0.9)}
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_when_get_next_instruction_called_then_expected_agamemnon_url_queried(
    mock_requests: MagicMock,
):
    configure_mock_agamemnon(mock_requests, None)
    _get_next_instruction("i03")
    mock_requests.get.assert_called_once_with(
        "http://agamemnon.diamond.ac.uk/getnextcollect/i03",
        headers={"Accept": "application/json"},
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_agamemnon_returns_an_unexpected_response_then_exception_is_thrown(
    mock_requests: MagicMock,
):
    mock_requests.get.return_value.content = json.dumps({"not_collect": ""})
    with pytest.raises(KeyError) as e:
        create_parameters_from_agamemnon()
    assert "not_collect" in str(e.value)


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_agamemnon_returns_multipin_when_get_next_pin_type_from_agamemnon_called_then_multipin_returned(
    mock_requests: MagicMock,
):
    configure_mock_agamemnon(mock_requests, "multipin_6x50+98.1")
    instruction, params = _instruction_and_data(_get_next_instruction("i03"))
    assert _get_pin_type_from_agamemnon_collect_parameters(params) == _PinType(
        6, 50, 98.1
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_leaves_parameters_unchanged_when_agamemnon_fails(
    mock_requests: MagicMock, load_centre_collect_params: LoadCentreCollect
):
    mock_requests.get.side_effect = Exception("Bad")
    old_grid_width = load_centre_collect_params.robot_load_then_centre.grid_width_um
    params = update_params_from_agamemnon(load_centre_collect_params)
    assert params.robot_load_then_centre.grid_width_um == old_grid_width


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.compare_params")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_changes_params_to_single_pin_when_agamemnon_gives_single_pin(
    mock_requests: MagicMock,
    mock_compare_params: MagicMock,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, None)
    load_centre_collect_params.robot_load_then_centre.grid_width_um = 0
    load_centre_collect_params.select_centres.n = 0
    params = update_params_from_agamemnon(load_centre_collect_params)
    assert (
        params.robot_load_then_centre.grid_width_um == GridscanParamConstants.WIDTH_UM
    )
    assert params.select_centres.n == 1
    assert params.multi_rotation_scan.snapshot_omegas_deg


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.compare_params")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_applies_multipin_attribs_given_agamemnon_returns_multipin(
    mock_requests: MagicMock,
    mock_compare_params: MagicMock,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, "multipin_6x50+10")
    params = update_params_from_agamemnon(load_centre_collect_params)
    assert params.robot_load_then_centre.grid_width_um == 270
    assert params.select_centres.n == 6
    assert params.robot_load_then_centre.tip_offset_um == 135
    assert not params.multi_rotation_scan.snapshot_omegas_deg


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_deduces_correct_url_given_set_of_parameters(
    mock_requests: MagicMock, load_centre_collect_params: LoadCentreCollect
):
    update_params_from_agamemnon(load_centre_collect_params)
    mock_requests.get.assert_called_once_with(
        "http://agamemnon.diamond.ac.uk/getnextcollect/i03",
        headers={"Accept": "application/json"},
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_logs_warning_when_exception_occurs(
    mock_requests: MagicMock,
    mock_logger: MagicMock,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, "multipin_unknown")

    update_params_from_agamemnon(load_centre_collect_params)

    mock_logger.warning.assert_called_once()
    assert (
        "Failed to update parameters: Agamemnon loop type of multipin_unknown "
        "not recognised" in mock_logger.warning.mock_calls[0].args[0]
    )


@pytest.mark.parametrize(
    "prefix, expected_visit",
    [
        ["/dls/i03/data/2025/mx23694-130/foo/bar", "mx23694-130"],
        ["/dls/not-i03/data/2021/mx84743-230", "mx84743-230"],
    ],
)
def test_given_valid_prefix_then_correct_visit_is_set(prefix: str, expected_visit: str):
    visit, _ = _get_withvisit_parameters_from_agamemnon(
        set_up_agamemnon_params(None, prefix, None)
    )
    assert visit == expected_visit


@pytest.mark.parametrize(
    "prefix",
    [
        "/not-dls/i03/data/2025/mx23694-130/foo/bar",
        "/dls/i03/not-data/2025/mx23694-130/foo/bar",
        "/foo/bar/i03/data/2025/mx23694-130",
    ],
)
def test_given_invalid_prefix_then_exception_raised(prefix: str):
    with pytest.raises(ValueError) as e:
        _get_withvisit_parameters_from_agamemnon(
            set_up_agamemnon_params(None, prefix, None)
        )

    assert "MX-General root structure" in str(e.value)


def test_no_prefix_raises_exception():
    with pytest.raises(KeyError) as e:
        _get_withvisit_parameters_from_agamemnon({"not_collect": ""})

    assert "Unexpected json from agamemnon" in str(e.value)


@pytest.mark.parametrize(
    "mock_error, mock_log",
    [
        (ValueError(), "Failed to compare parameters: "),
        (Exception(), "Unexpected error occurred. Failed to compare parameters: "),
    ],
)
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch(
    "mx_bluesky.hyperion.external_interaction.agamemnon._populate_parameters_from_agamemnon"
)
def test_compare_params_logs_exception_if_fails_to_populate_parameters_from_hyperion(
    mock_populate_params,
    mock_logger,
    mock_requests,
    mock_error,
    mock_log,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, None)
    mock_populate_params.side_effect = mock_error
    compare_params(
        load_centre_collect_params,
    )
    assert mock_log in mock_logger.mock_calls[0][1][0]


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch(
    "mx_bluesky.hyperion.external_interaction.agamemnon.create_parameters_from_agamemnon",
    MagicMock(return_value=[]),
)
def test_compare_params_logs_message_if_agamemnon_returns_no_instructions(
    mock_logger: MagicMock, load_centre_collect_params: LoadCentreCollect
):
    compare_params(
        load_centre_collect_params,
    )
    mock_logger.info.assert_called_once_with("Agamemnon returned no instructions")


@pytest.fixture
def agamemnon_response(request) -> Generator[str, None, None]:
    with (
        patch("mx_bluesky.common.parameters.components.os", new=MagicMock()),
        patch(
            "mx_bluesky.hyperion.external_interaction.agamemnon.requests"
        ) as mock_requests,
        open(request.param) as json_file,
    ):
        example_json = json_file.read()
        mock_requests.get.return_value.content = example_json
        yield example_json


@pytest.mark.parametrize(
    "agamemnon_response",
    [
        "tests/test_data/agamemnon/example_native.json",
        "tests/test_data/agamemnon/example_collect_multipin.json",
    ],
    indirect=True,
)
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch("mx_bluesky.common.parameters.rotation.os", new=MagicMock())
@patch("dodal.devices.detector.detector.Path", new=MagicMock())
@patch("dodal.utils.os", new=MagicMock())
def test_compare_params_causes_no_warning_when_compared_to_gda_params(
    mock_logger: MagicMock,
    agamemnon_response: str,
    load_centre_collect_params: LoadCentreCollect,
):
    compare_params(load_centre_collect_params)
    mock_logger.warning.assert_not_called()


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_contains_expected_data(agamemnon_response):
    hyperion_params_list = create_parameters_from_agamemnon()
    for hyperion_params in hyperion_params_list:
        assert isinstance(hyperion_params, LoadCentreCollectParams)
        assert hyperion_params.visit == "mx34598-77"
        assert isclose(hyperion_params.detector_distance_mm, 237.017, abs_tol=1e-3)  # type: ignore
        assert hyperion_params.sample_id == 6501159
        assert hyperion_params.sample_puck == 5
        assert hyperion_params.sample_pin == 4
        assert hyperion_params.select_centres.n == 1


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_contains_expected_robot_load_then_centre_data(
    agamemnon_response,
):
    hyperion_params_list = create_parameters_from_agamemnon()
    load_centre_collect_list = [
        load_centre_collect_to_internal(p)
        for p in hyperion_params_list
        if isinstance(p, LoadCentreCollectParams)
    ]
    assert len(hyperion_params_list) == len(load_centre_collect_list) == 2

    assert load_centre_collect_list[0].robot_load_then_centre.chi_start_deg == 0.0
    assert load_centre_collect_list[1].robot_load_then_centre.chi_start_deg == 30.0
    for robot_load_params in [
        params.robot_load_then_centre for params in load_centre_collect_list
    ]:
        assert robot_load_params.visit == "mx34598-77"
        assert isclose(robot_load_params.detector_distance_mm, 237.017, abs_tol=1e-3)  # type: ignore
        assert robot_load_params.sample_id == 6501159
        assert robot_load_params.sample_puck == 5
        assert robot_load_params.sample_pin == 4
        assert robot_load_params.demand_energy_ev == 12700.045934258673
        assert robot_load_params.omega_start_deg == 0.0
        assert robot_load_params.transmission_frac == 1.0
        assert robot_load_params.tip_offset_um == 300.0
        assert robot_load_params.grid_width_um == 600.0
        assert str(robot_load_params.parameter_model_version) == "5.3.0"
        assert (
            robot_load_params.storage_directory
            == "/dls/i03/data/2025/mx34598-77/auto/CBLBA/CBLBA-x00242/xraycentring"
        )
        assert robot_load_params.file_name == "CBLBA-x00242"
        assert robot_load_params.snapshot_directory == PosixPath(
            "/dls/i03/data/2025/mx34598-77/auto/CBLBA/CBLBA-x00242/xraycentring/snapshots"
        )


@patch("mx_bluesky.common.parameters.rotation.os", new=MagicMock())
@patch("dodal.devices.detector.detector.Path", new=MagicMock())
@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_contains_expected_rotation_data(
    agamemnon_response,
):
    hyperion_params_list = [
        load_centre_collect_to_internal(cast(LoadCentreCollectParams, p))
        for p in create_parameters_from_agamemnon()
    ]  # type: ignore
    assert len(hyperion_params_list) == 2
    for hyperion_params in hyperion_params_list:
        rotation_params = hyperion_params.multi_rotation_scan
        assert rotation_params.visit == "mx34598-77"
        assert isclose(rotation_params.detector_distance_mm, 237.017, abs_tol=1e-3)  # type: ignore
        assert rotation_params.detector_params.omega_start == 0.0
        assert rotation_params.detector_params.exposure_time_s == 0.003
        assert rotation_params.detector_params.num_images_per_trigger == 3600
        assert rotation_params.num_images == 3600
        assert rotation_params.transmission_frac == 0.1426315789473684
        assert rotation_params.comment == "Complete_P1_sweep1 "
        assert rotation_params.ispyb_experiment_type == "OSC"

        assert rotation_params.demand_energy_ev == 12700.045934258673
        assert str(rotation_params.parameter_model_version) == "5.3.0"
        assert (
            rotation_params.storage_directory
            == "/dls/i03/data/2025/mx34598-77/auto/CBLBA/CBLBA-x00242"
        )
        assert rotation_params.file_name == "CBLBA-x00242"
        assert rotation_params.snapshot_directory == PosixPath(
            "/dls/i03/data/2025/mx34598-77/auto/CBLBA/CBLBA-x00242/snapshots"
        )

    individual_scans = list(
        hyperion_params_list[0].multi_rotation_scan.single_rotation_scans  # type: ignore
    ) + list(
        hyperion_params_list[1].multi_rotation_scan.single_rotation_scans  # type: ignore
    )
    assert len(individual_scans) == 2
    assert individual_scans[0].scan_points["omega"][1] == 0.1
    assert individual_scans[0].phi_start_deg == 0.0
    assert individual_scans[0].chi_start_deg == 0.0
    assert individual_scans[0].rotation_direction == RotationDirection.POSITIVE
    assert individual_scans[1].scan_points["omega"][1] == 0.1
    assert individual_scans[1].phi_start_deg == 0.0
    assert individual_scans[1].chi_start_deg == 30.0
    assert individual_scans[1].rotation_direction == RotationDirection.POSITIVE


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_collect_multipin.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_populates_multipin_parameters_from_agamemnon(
    agamemnon_response,
):
    hyperion_params_list = create_parameters_from_agamemnon()
    for hyperion_params in hyperion_params_list:
        assert hyperion_params.select_centres.n == 6  # type: ignore


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_creates_multiple_load_centre_collect_for_native_collection(
    agamemnon_response,
):
    hyperion_params_list = create_parameters_from_agamemnon()
    assert len(hyperion_params_list) == 2
    assert (
        sum(
            [
                len(hyperion_params.multi_rotation_scan.rotation_scans)  # type: ignore
                for hyperion_params in hyperion_params_list
            ]
        )
        == 2
    )


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_get_withenergy_parameters_from_agamemnon(agamemnon_response):
    _, agamemnon_params = _instruction_and_data(_get_next_instruction("i03"))
    demand_energy_ev = _get_withenergy_parameters_from_agamemnon(agamemnon_params)
    assert demand_energy_ev["demand_energy_ev"] == 12700.045934258673


def test_get_withenergy_parameters_from_agamemnon_when_no_wavelength():
    agamemnon_params = {}
    demand_energy_ev = _get_withenergy_parameters_from_agamemnon(agamemnon_params)
    assert demand_energy_ev["demand_energy_ev"] is None


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_create_parameters_from_agamemnon_returns_empty_list_if_collect_instruction_is_empty(
    mock_requests,
):
    mock_requests.get.return_value.content = json.dumps({"collect": {}})
    params = create_parameters_from_agamemnon()
    assert params == []


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_create_parameters_from_agamemnon_returns_empty_list_if_no_instruction(
    mock_requests,
):
    mock_requests.get.return_value.content = json.dumps({})
    params = create_parameters_from_agamemnon()
    assert params == []


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_collect_multipin.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_does_not_return_none_if_queue_is_not_empty(
    agamemnon_response,
):
    params = create_parameters_from_agamemnon()
    assert params is not None


@pytest.mark.parametrize(
    "agamemnon_response", ["tests/test_data/agamemnon/example_wait.json"], indirect=True
)
def test_create_parameters_from_agamemnon_creates_wait(agamemnon_response):
    params = create_parameters_from_agamemnon()
    assert len(params) == 1
    assert isinstance(params[0], Wait)
    assert params[0].duration_s == 12.34



from unittest.mock import AsyncMock, MagicMock, patch

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pytest
from bluesky.callbacks import CallbackBase
from bluesky.run_engine import RunEngine
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from ophyd_async.core import completed_status
from ophyd_async.fastcs.jungfrau import (
    AcquisitionType,
    GainMode,
    PedestalMode,
)

from mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.do_darks import (
    do_pedestal_darks,
)


class CheckMonitor(CallbackBase):
    """Store the order and values of updates to specified signals

    Usage: Instantiate this callback with list of signals to track, and subscribe the run_engine to this
    callback. Run your plan using Bluesky's monitor_during decorator or wrapper, specifing the same signals
    in the monitor.
    """

    def __init__(self, signals_to_track: list[str]):
        self.signals_and_values = {signal: [] for signal in signals_to_track}

    def event(self, doc):
        key, value = next(iter(doc["data"].items()))
        self.signals_and_values[key].append(value)
        return doc


def fake_complete(_, group=None):
    yield from bps.null()
    return completed_status()


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils.log_on_percentage_complete",
    new=MagicMock,
)
@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils.bps.complete",
    new=MagicMock(side_effect=fake_complete),
)
async def test_full_do_pedestal_darks(
    jungfrau: CommissioningJungfrau,
    run_engine: RunEngine,
):
    # Test that plan succeeds in RunEngine and pedestal-specific signals are changed as expected

    @bpp.run_decorator()
    def test_plan():
        yield from bps.monitor(jungfrau.drv.acquisition_type, name="AT")
        yield from bps.monitor(jungfrau.drv.pedestal_mode_state, name="PM")
        yield from bps.monitor(jungfrau.drv.gain_mode, name="GM")
        yield from do_pedestal_darks(0.001, 2, 2, jungfrau=jungfrau)

    jungfrau._controller.arm = AsyncMock()
    assert await jungfrau.drv.acquisition_type.get_value() == AcquisitionType.STANDARD
    await jungfrau.drv.gain_mode.set(GainMode.FIX_G2)
    await jungfrau.drv.pedestal_mode_state.set(PedestalMode.OFF)
    monitor_tracker = CheckMonitor(
        [
            "detector-drv-acquisition_type",
            "detector-drv-pedestal_mode_state",
            "detector-drv-gain_mode",
        ]
    )
    run_engine.subscribe(monitor_tracker)
    run_engine(test_plan())

    assert monitor_tracker.signals_and_values["detector-drv-acquisition_type"] == [
        AcquisitionType.STANDARD,  # Repeated as staging JF also sets to standard
        AcquisitionType.STANDARD,
        AcquisitionType.PEDESTAL,
        AcquisitionType.STANDARD,
    ]
    assert monitor_tracker.signals_and_values["detector-drv-pedestal_mode_state"] == [
        PedestalMode.OFF,  # Repeated as staging JF also turns pedestals off
        PedestalMode.OFF,
        PedestalMode.ON,
        PedestalMode.OFF,
    ]

    # When using the real detector, the switching of gain mode is a bit more complicated,
    # see the docstring for the do_pedestal_darks plan.
    assert monitor_tracker.signals_and_values["detector-drv-gain_mode"] == [
        GainMode.FIX_G2,
        GainMode.DYNAMIC,
    ]


class FakeError(Exception): ...


async def test_jungfrau_unstage_on_error(
    jungfrau: CommissioningJungfrau, run_engine: RunEngine
):
    jungfrau.stage = MagicMock(side_effect=FakeError)
    jungfrau.unstage = MagicMock(side_effect=lambda: completed_status())

    def test_plan():
        yield from do_pedestal_darks(0.001, 2, 2, jungfrau=jungfrau)

    with pytest.raises(FakeError):
        run_engine(test_plan())
    assert jungfrau.unstage.call_count == 1



from __future__ import annotations

from datetime import datetime
from pathlib import Path

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pydantic
from dodal.devices.motors import XYZOmegaStage, XYZStage
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot, SampleLocation

from mx_bluesky.beamlines.aithre_lasershaping.parameters.constants import CONST
from mx_bluesky.beamlines.aithre_lasershaping.parameters.robot_load_parameters import (
    AithreRobotLoad,
)
from mx_bluesky.common.device_setup_plans.robot_load_unload import (
    do_plan_while_lower_gonio_at_home,
)
from mx_bluesky.common.experiment_plans.pin_tip_centring_plan import (
    PinTipCentringComposite,
    pin_tip_centre_plan,
)
from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
    PlanNameConstants,
)


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class RobotLoadComposite:
    # RobotLoad fields
    robot: BartRobot
    lower_gonio: XYZStage
    oav: OAV
    gonio: XYZOmegaStage


def _move_gonio_to_home_position(
    composite: RobotLoadComposite,
    x_home: float = 0.0,
    y_home: float = 0.0,
    z_home: float = 0.0,
    omega_home: float = 0.0,
    group: str = "group",
):
    """
    Move Gonio to home position, default is zero
    """
    yield from bps.abs_set(composite.gonio.omega, omega_home, group=group)
    yield from bps.abs_set(composite.gonio.x, x_home, group=group)
    yield from bps.abs_set(composite.gonio.y, y_home, group=group)
    yield from bps.abs_set(composite.gonio.z, z_home, group=group)

    yield from bps.wait(group=group)


def _take_robot_snapshots(oav: OAV, directory: Path):
    time_now = datetime.now()
    snapshot_format = f"{time_now.strftime('%H%M%S')}_{{device}}_after_load"
    for device in [oav.snapshot]:
        yield from bps.abs_set(
            device.filename, snapshot_format.format(device=device.name)
        )
        yield from bps.abs_set(device.directory, str(directory))
        # Note: should be able to use `wait=True` after https://github.com/bluesky/bluesky/issues/1795
        yield from bps.trigger(device, group="snapshots")
        yield from bps.wait("snapshots")


def _do_robot_load_and_centre(
    composite: RobotLoadComposite,
    sample_location: SampleLocation,
    sample_id: int,
    pin_tip_detection: PinTipDetection,
    tip_offset_microns: float = 0,
    oav_config_file: str = CONST.OAV_CENTRING_FILE,
):
    yield from bps.abs_set(composite.robot.next_sample_id, sample_id, wait=True)
    yield from bps.abs_set(
        composite.robot,
        sample_location,
        group="robot_load",
    )

    yield from _move_gonio_to_home_position(composite=composite, group="robot_load")

    yield from bps.wait(group="robot_load")

    pin_tip_centring_composite = PinTipCentringComposite(
        composite.oav, composite.gonio, pin_tip_detection
    )
    yield from pin_tip_centre_plan(
        pin_tip_centring_composite, tip_offset_microns, oav_config_file
    )


def _robot_load_and_snapshots(
    composite: RobotLoadComposite,
    location: SampleLocation,
    snapshot_directory: Path,
    sample_id: int,
    pin_tip_detection: PinTipDetection,
    tip_offset_microns: float = 0,
    oav_config_file: str = CONST.OAV_CENTRING_FILE,
):
    yield from bps.create(name=DocDescriptorNames.ROBOT_PRE_LOAD)
    yield from bps.read(composite.robot)
    yield from bps.save()

    robot_load_plan = _do_robot_load_and_centre(
        composite,
        location,
        sample_id,
        pin_tip_detection,
        tip_offset_microns,
        oav_config_file,
    )

    gonio_finished = yield from do_plan_while_lower_gonio_at_home(
        robot_load_plan, composite.lower_gonio
    )
    yield from bps.wait(group="snapshot")

    yield from _take_robot_snapshots(composite.oav, snapshot_directory)

    yield from bps.create(name=DocDescriptorNames.ROBOT_UPDATE)
    yield from bps.read(composite.robot)
    yield from bps.read(composite.oav.snapshot)
    yield from bps.save()

    yield from bps.wait(gonio_finished)


def robot_load_and_snapshots_plan(
    composite: RobotLoadComposite,
    params: AithreRobotLoad,
    ptd: PinTipDetection,
    tip_offset_microns: float = 0,
    oav_config_file: str = CONST.OAV_CENTRING_FILE,
):
    assert params.sample_puck is not None
    assert params.sample_pin is not None

    sample_location = SampleLocation(params.sample_puck, params.sample_pin)

    yield from _move_gonio_to_home_position(composite)

    yield from bpp.set_run_key_wrapper(
        bpp.run_wrapper(
            _robot_load_and_snapshots(
                composite,
                sample_location,
                params.snapshot_directory,
                params.sample_id,
                ptd,
                tip_offset_microns,
                oav_config_file,
            ),
            md={
                "subplan_name": PlanNameConstants.ROBOT_LOAD,
                "metadata": {"visit": params.visit, "sample_id": params.sample_id},
                "activate_callbacks": [
                    "RobotLoadISPyBCallback",
                ],
            },
        ),
        PlanNameConstants.ROBOT_LOAD,
    )


def robot_unload_plan(
    composite: RobotLoadComposite,
    params: AithreRobotLoad,
):
    @bpp.run_decorator(
        md={
            "subplan_name": PlanNameConstants.ROBOT_UNLOAD,
            "metadata": {"visit": params.visit, "sample_id": params.sample_id},
            "activate_callbacks": [
                "RobotLoadISPyBCallback",
            ],
        },
    )
    def do_robot_unload_and_send_to_ispyb():
        yield from _take_robot_snapshots(composite.oav, params.snapshot_directory)
        yield from bps.wait(group="snapshot")
        yield from _move_gonio_to_home_position(composite)

        yield from bps.abs_set(composite.robot, None, wait=True)

        yield from bps.create(name=DocDescriptorNames.ROBOT_UPDATE)
        yield from bps.read(composite.robot)
        yield from bps.read(composite.oav.snapshot)
        yield from bps.save()

    yield from do_robot_unload_and_send_to_ispyb()



import dataclasses
import os
import re
from collections.abc import Iterator
from datetime import datetime
from math import cos, radians, sin
from pathlib import Path

from dodal.devices.oav.snapshots.snapshot_image_processing import (
    compute_beam_centre_pixel_xy_for_mm_position,
    draw_crosshair,
)
from event_model import Event, EventDescriptor, RunStart
from PIL import Image

from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)
from mx_bluesky.common.parameters.components import WithSnapshot
from mx_bluesky.common.parameters.constants import DocDescriptorNames, PlanNameConstants
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER as CALLBACK_LOGGER

COMPRESSION_LEVEL = 6  # 6 is the default compression level for PIL if not specified


@dataclasses.dataclass
class _SnapshotInfo:
    beam_centre: tuple[int, int]
    microns_per_pixel: tuple[float, float]
    snapshot_path: str
    omega: int
    sample_pos_mm: tuple[float, float, float]

    @property
    def snapshot_basename(self) -> str:
        match = re.match("(.*)\\.png", self.snapshot_path)
        assert match, f"Snapshot {self.snapshot_path} was not a .png file"
        return match.groups()[0]


class BeamDrawingCallback(PlanReactiveCallback):
    """
    Callback that monitors for OAV_ROTATION_SNAPSHOT_TRIGGERED events and
    draws a crosshair at the beam centre, saving the snapshot to a file.
    The callback assumes an OAV device "oav" and Smargon "gonio"
    Examples:
        Take a rotation snapshot at the current location
    >>> from bluesky.run_engine import RunEngine
    >>> import bluesky.preprocessors as bpp
    >>> import bluesky.plan_stubs as bps
    >>> from dodal.devices.oav.oav_detector import OAV
    >>> from mx_bluesky.common.parameters.components import WithSnapshot
    >>> def take_snapshot(params: WithSnapshot, oav: OAV, run_engine: RunEngine):
    ...     run_engine.subscribe(BeamDrawingCallback())
    ...     @bpp.run_decorator(md={
    ...     "activate_callbacks": ["BeamDrawingCallback"],
    ...         "with_snapshot": params.model_dump_json(),
    ...     })
    ...     def inner_plan():
    ...         yield from bps.abs_set(oav.snapshot.directory, "/path/to/snapshot_folder", wait=True)
    ...         yield from bps.abs_set(oav.snapshot.filename, "my_snapshot_prefix", wait=True)
    ...         yield from bps.trigger(oav.snapshot, wait=True)
    ...         yield from bps.create(DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED)
    ...         yield from bps.read(oav)
    ...         yield from bps.save()

        Generate rotation snapshots from a previously taken base gridscan snapshot.
        WithSnapshot.snapshot_omegas_deg is ignored and snapshots are generated for the previously captured
        0, 90 base images named "my_snapshot_prefix_0" and "my_snapshot_prefix_90"
    >>> from dodal.devices.smargon import Smargon
    >>> def take_snapshot(params: WithSnapshot, oav: OAV, smargon: Smargon, run_engine: RunEngine):
    ...     run_engine.subscribe(BeamDrawingCallback())
    ...     @bpp.run_decorator(md={
    ...     "activate_callbacks": ["BeamDrawingCallback"],
    ...         "with_snapshot": params.model_dump_json(),
    ...     })
    ...     def inner_plan():
    ...         for omega in (0, 90,):
    ...             yield from bps.abs_set(smargon.omega, omega, wait=True)
    ...             yield from bps.abs_set(oav.grid_snapshot.directory, "/path/to/grid_snapshot_folder", wait=True)
    ...             yield from bps.abs_set(oav.grid_snapshot.filename, f"my_grid_snapshot_prefix_{omega}", wait=True)
    ...             yield from bps.trigger(oav.grid_snapshot, wait=True)
    ...             yield from bps.create(DocDescriptorNames.OAV_GRID_SNAPSHOT_TRIGGERED)
    ...             yield from bps.read(oav)        # Capture base image path
    ...             yield from bps.read(smargon)    # Capture base image sample x, y, z, omega
    ...             yield from bps.save()
    ...             # Rest of gridscan here...
    ...         # Later on...
    ...         for omega in (0, 90,):
    ...             yield from bps.abs_set(oav.snapshot.last_saved_path,
    ...                 f"/path/to/snapshot_folder/my_snapshot_prefix_{omega}.png", wait=True)
    ...             yield from bps.create(DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED)
    ...             yield from bps.read(oav)            # Capture path info for generated snapshot
    ...             yield from bps.read(smargon)        # Capture the current sample x, y, z
    ...             yield from bps.save()
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, log=CALLBACK_LOGGER, **kwargs)
        self._base_snapshots: list[_SnapshotInfo] = []
        self._rotation_snapshot_descriptor: str = ""
        self._grid_snapshot_descriptor: str = ""
        self._next_snapshot_info: Iterator | None = None
        self._use_grid_snapshots: bool = False

    def _reset(self):
        self._base_snapshots = []

    def activity_gated_start(self, doc: RunStart):
        if self.activity_uid == doc.get("uid"):
            self._reset()
            with_snapshot = WithSnapshot.model_validate_json(doc.get("with_snapshot"))  # type: ignore
            self._use_grid_snapshots = with_snapshot.use_grid_snapshots
            CALLBACK_LOGGER.info(f"Snapshot callback initialised with {with_snapshot}")
        elif doc.get("subplan_name") == PlanNameConstants.ROTATION_MAIN:
            self._next_snapshot_info = None
            CALLBACK_LOGGER.info("Snapshot callback start rotation")
        return doc

    def activity_gated_descriptor(self, doc: EventDescriptor) -> EventDescriptor | None:
        if doc.get("name") == DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED:
            self._rotation_snapshot_descriptor = doc["uid"]
        elif doc.get("name") == DocDescriptorNames.OAV_GRID_SNAPSHOT_TRIGGERED:
            self._grid_snapshot_descriptor = doc["uid"]
        return doc

    def activity_gated_event(self, doc: Event) -> Event:
        if doc["descriptor"] == self._rotation_snapshot_descriptor:
            self._handle_rotation_snapshot(doc)
        elif doc["descriptor"] == self._grid_snapshot_descriptor:
            self._handle_grid_snapshot(doc)
        return doc

    def _extract_base_snapshot_params(
        self, snapshot_device_prefix: str, doc: Event
    ) -> _SnapshotInfo:
        data = doc["data"]
        base_snapshot_path = data[f"oav-{snapshot_device_prefix}-last_saved_path"]
        return _SnapshotInfo(
            beam_centre=(data["oav-beam_centre_i"], data["oav-beam_centre_j"]),
            microns_per_pixel=(
                data["oav-microns_per_pixel_x"],
                data["oav-microns_per_pixel_y"],
            ),
            snapshot_path=base_snapshot_path,
            sample_pos_mm=(
                data.get("gonio-x", 0.0),
                data.get("gonio-y", 0.0),
                data.get("gonio-z", 0.0),
            ),
            omega=round(data.get("gonio-omega", 0.0)),
        )

    def _handle_grid_snapshot(self, doc: Event):
        snapshot_info = self._extract_base_snapshot_params("grid_snapshot", doc)
        self._base_snapshots.append(snapshot_info)

    def _handle_rotation_snapshot(self, doc: Event) -> Event:
        data = doc["data"]
        if self._use_grid_snapshots:
            if not self._next_snapshot_info:
                self._next_snapshot_info = iter(self._base_snapshots)
            snapshot_info = next(self._next_snapshot_info, None)
            assert snapshot_info, (
                "Insufficient base gridscan snapshots to generate required rotation snapshots"
            )
            current_sample_pos_mm = (
                data["gonio-x"],
                data["gonio-y"],
                data["gonio-z"],
            )
            CALLBACK_LOGGER.info(
                f"Generating snapshot at {current_sample_pos_mm} from base snapshot {snapshot_info}"
            )
            output_snapshot_directory = data["oav-snapshot-directory"]
            if not os.path.exists(output_snapshot_directory):
                os.mkdir(output_snapshot_directory)
            base_file_stem = Path(snapshot_info.snapshot_path).stem
            output_snapshot_filename = _snapshot_filename(base_file_stem)
            output_snapshot_path = (
                f"{output_snapshot_directory}/{output_snapshot_filename}.png"
            )
            self._generate_snapshot_at(
                snapshot_info,
                output_snapshot_path,
                *self._image_plane_offset_mm(snapshot_info, current_sample_pos_mm),
            )
        else:
            snapshot_info = self._extract_base_snapshot_params("snapshot", doc)
            output_snapshot_path = (
                f"{snapshot_info.snapshot_basename}_with_beam_centre.png"
            )
            CALLBACK_LOGGER.info(
                f"Annotating snapshot {output_snapshot_path} from base snapshot {snapshot_info}"
            )
            self._generate_snapshot_zero_offset(
                snapshot_info,
                output_snapshot_path,
            )
        data["oav-snapshot-last_saved_path"] = output_snapshot_path
        return doc

    def _image_plane_offset_mm(
        self,
        snapshot_info: _SnapshotInfo,
        current_sample_pos_mm: tuple[float, float, float],
    ) -> tuple[float, float]:
        return self._project_xyz_to_xy(
            (
                (current_sample_pos_mm[0] - snapshot_info.sample_pos_mm[0]),
                (current_sample_pos_mm[1] - snapshot_info.sample_pos_mm[1]),
                (current_sample_pos_mm[2] - snapshot_info.sample_pos_mm[2]),
            ),
            snapshot_info.omega,
        )

    def _project_xyz_to_xy(
        self, xyz: tuple[float, float, float], omega_deg: float
    ) -> tuple[float, float]:
        return (
            xyz[0],
            xyz[1] * cos(-radians(omega_deg)) + xyz[2] * sin(-radians(omega_deg)),
        )

    def _generate_snapshot_zero_offset(
        self,
        base_snapshot_info: _SnapshotInfo,
        output_snapshot_path: str,
    ):
        self._generate_snapshot_at(base_snapshot_info, output_snapshot_path, 0, 0)

    def _generate_snapshot_at(
        self,
        base_snapshot_info: _SnapshotInfo,
        output_snapshot_path: str,
        image_plane_dx_mm: float,
        image_plane_dy_mm: float,
    ):
        """
        Save a snapshot to the specified path, with an annotated crosshair at the specified
        position
        Args:
            base_snapshot_info: Metadata about the base snapshot image from which the annotated
                image will be derived.
            output_snapshot_path:  The path to the image that will be annotated.
            image_plane_dx_mm: Relative x location of the sample to the original image in the image plane (mm)
            image_plane_dy_mm: Relative y location of the sample to the original image in the image plane (mm)
        """
        image = Image.open(base_snapshot_info.snapshot_path)
        x_px, y_px = compute_beam_centre_pixel_xy_for_mm_position(
            (image_plane_dx_mm, image_plane_dy_mm),
            base_snapshot_info.beam_centre,
            base_snapshot_info.microns_per_pixel,
        )
        draw_crosshair(image, x_px, y_px)
        image.save(output_snapshot_path, format="png", compress_level=COMPRESSION_LEVEL)


def _snapshot_filename(grid_snapshot_name):
    time_now = datetime.now()
    filename = f"{time_now.strftime('%H%M%S%f')[:8]}_oav_snapshot_{grid_snapshot_name}"
    return filename
